@@ -130,12 +130,13 @@ buf_init_chartab(buf, global)
      */
     vim_memset(buf->b_chartab, 0, (size_t)32);
 #ifdef FEAT_MBYTE
-    for (c = 0; c < 256; ++c)
-    {
-	/* double-byte characters are probably word characters */
-	if (enc_dbcs != 0 && MB_BYTE2LEN(c) == 2)
-	    SET_CHARTAB(buf, c);
-    }
+    if (enc_dbcs != 0)
+	for (c = 0; c < 256; ++c)
+	{
+	    /* double-byte characters are probably word characters */
+	    if (MB_BYTE2LEN(c) == 2)
+		SET_CHARTAB(buf, c);
+	}
 #endif
 
 #ifdef FEAT_LISP
@@ -170,7 +171,7 @@ buf_init_chartab(buf, global)
 		tilde = TRUE;
 		++p;
 	    }
-	    if (isdigit(*p))
+	    if (VIM_ISDIGIT(*p))
 		c = getdigits(&p);
 	    else
 		c = *p++;
@@ -178,7 +179,7 @@ buf_init_chartab(buf, global)
 	    if (*p == '-' && p[1] != NUL)
 	    {
 		++p;
-		if (isdigit(*p))
+		if (VIM_ISDIGIT(*p))
 		    c2 = getdigits(&p);
 		else
 		    c2 = *p++;
@@ -206,7 +207,10 @@ buf_init_chartab(buf, global)
 	    }
 	    while (c <= c2)
 	    {
-		if (!do_isalpha || isalpha(c)
+		/* Use the MB_ functions here, because isalpha() doesn't
+		 * work properly when 'encoding' is "latin1" and the locale is
+		 * "C".  */
+		if (!do_isalpha || MB_ISLOWER(c) || MB_ISUPPER(c)
 #ifdef FEAT_FKMAP
 			|| (p_altkeymap && (F_isalpha(c) || F_isdigit(c)))
 #endif
@@ -274,8 +278,6 @@ buf_init_chartab(buf, global)
     return OK;
 }
 
-#if defined(FEAT_TITLE) || defined(FEAT_STL_OPT) || defined(FEAT_WINDOWS) \
-	|| defined(PROTO)
 /*
  * Translate any special characters in buf[bufsize] in-place.
  * The result is a string with only printable characters, but if there is not
@@ -297,7 +299,7 @@ trans_characters(buf, bufsize)
     {
 # ifdef FEAT_MBYTE
 	/* Assume a multi-byte character doesn't need translation. */
-	if (has_mbyte && (trs_len = (*mb_ptr2len_check)(buf)) > 1)
+	if (has_mbyte && (trs_len = (*mb_ptr2len)(buf)) > 1)
 	    len -= trs_len;
 	else
 # endif
@@ -317,9 +319,9 @@ trans_characters(buf, bufsize)
 	buf += trs_len;
     }
 }
-#endif
 
-#if defined(FEAT_EVAL) || defined(FEAT_TITLE) || defined(PROTO)
+#if defined(FEAT_EVAL) || defined(FEAT_TITLE) || defined(FEAT_INS_EXPAND) \
+	|| defined(PROTO)
 /*
  * Translate a string into allocated memory, replacing special chars with
  * printable chars.  Returns NULL when out of memory.
@@ -331,22 +333,30 @@ transstr(s)
     char_u	*res;
     char_u	*p;
 #ifdef FEAT_MBYTE
-    int		l, len;
+    int		l, len, c;
+    char_u	hexbuf[11];
 #endif
 
 #ifdef FEAT_MBYTE
     if (has_mbyte)
     {
-	/* Compute the length of the result, taking into account that
-	 * multi-byte characters are copied unchanged. */
+	/* Compute the length of the result, taking account of unprintable
+	 * multi-byte characters. */
 	len = 0;
 	p = s;
 	while (*p != NUL)
 	{
-	    if ((l = (*mb_ptr2len_check)(p)) > 1)
+	    if ((l = (*mb_ptr2len)(p)) > 1)
 	    {
-		len += l;
+		c = (*mb_ptr2char)(p);
 		p += l;
+		if (vim_isprintc(c))
+		    len += l;
+		else
+		{
+		    transchar_hex(hexbuf, c);
+		    len += (int)STRLEN(hexbuf);
+		}
 	    }
 	    else
 	    {
@@ -369,9 +379,13 @@ transstr(s)
 	while (*p != NUL)
 	{
 #ifdef FEAT_MBYTE
-	    if (has_mbyte && (l = (*mb_ptr2len_check)(p)) > 1)
+	    if (has_mbyte && (l = (*mb_ptr2len)(p)) > 1)
 	    {
-		STRNCAT(res, p, l);	/* append printable multi-byte char */
+		c = (*mb_ptr2char)(p);
+		if (vim_isprintc(c))
+		    STRNCAT(res, p, l);	/* append printable multi-byte char */
+		else
+		    transchar_hex(res + STRLEN(res), c);
 		p += l;
 	    }
 	    else
@@ -385,36 +399,115 @@ transstr(s)
 
 #if defined(FEAT_SYN_HL) || defined(FEAT_INS_EXPAND) || defined(PROTO)
 /*
- * Convert the string "p" to do ignore-case comparing.
- * It's done in-place.
+ * Convert the string "str[orglen]" to do ignore-case comparing.  Uses the
+ * current locale.
+ * When "buf" is NULL returns an allocated string (NULL for out-of-memory).
+ * Otherwise puts the result in "buf[buflen]".
  */
-    void
-str_foldcase(p)
-    char_u	*p;
+    char_u *
+str_foldcase(str, orglen, buf, buflen)
+    char_u	*str;
+    int		orglen;
+    char_u	*buf;
+    int		buflen;
 {
-    while (*p != NUL)
+    garray_T	ga;
+    int		i;
+    int		len = orglen;
+
+#define GA_CHAR(i)  ((char_u *)ga.ga_data)[i]
+#define GA_PTR(i)   ((char_u *)ga.ga_data + i)
+#define STR_CHAR(i)  (buf == NULL ? GA_CHAR(i) : buf[i])
+#define STR_PTR(i)   (buf == NULL ? GA_PTR(i) : buf + i)
+
+    /* Copy "str" into "buf" or allocated memory, unmodified. */
+    if (buf == NULL)
+    {
+	ga_init2(&ga, 1, 10);
+	if (ga_grow(&ga, len + 1) == FAIL)
+	    return NULL;
+	mch_memmove(ga.ga_data, str, (size_t)len);
+	ga.ga_len = len;
+    }
+    else
+    {
+	if (len >= buflen)	    /* Ugly! */
+	    len = buflen - 1;
+	mch_memmove(buf, str, (size_t)len);
+    }
+    if (buf == NULL)
+	GA_CHAR(len) = NUL;
+    else
+	buf[len] = NUL;
+
+    /* Make each character lower case. */
+    i = 0;
+    while (STR_CHAR(i) != NUL)
     {
 #ifdef FEAT_MBYTE
-	if (has_mbyte && MB_BYTE2LEN(*p) > 1)
+	if (enc_utf8 || (has_mbyte && MB_BYTE2LEN(STR_CHAR(i)) > 1))
 	{
 	    if (enc_utf8)
 	    {
 		int	c, lc;
 
-		c = utf_ptr2char(p);
+		c = utf_ptr2char(STR_PTR(i));
 		lc = utf_tolower(c);
-		if (c != lc && utf_char2len(c) == utf_char2len(lc))
-		    (void)utf_char2bytes(c, p);
+		if (c != lc)
+		{
+		    int	    ol = utf_char2len(c);
+		    int	    nl = utf_char2len(lc);
+
+		    /* If the byte length changes need to shift the following
+		     * characters forward or backward. */
+		    if (ol != nl)
+		    {
+			if (nl > ol)
+			{
+			    if (buf == NULL ? ga_grow(&ga, nl - ol + 1) == FAIL
+						    : len + nl - ol >= buflen)
+			    {
+				/* out of memory, keep old char */
+				lc = c;
+				nl = ol;
+			    }
+			}
+			if (ol != nl)
+			{
+			    if (buf == NULL)
+			    {
+				mch_memmove(GA_PTR(i) + nl, GA_PTR(i) + ol,
+						  STRLEN(GA_PTR(i) + ol) + 1);
+				ga.ga_len += nl - ol;
+			    }
+			    else
+			    {
+				mch_memmove(buf + i + nl, buf + i + ol,
+						    STRLEN(buf + i + ol) + 1);
+				len += nl - ol;
+			    }
+			}
+		    }
+		    (void)utf_char2bytes(lc, STR_PTR(i));
+		}
 	    }
-	    p += (*mb_ptr2len_check)(p);	/* skip multi-byte char */
+	    /* skip to next multi-byte char */
+	    i += (*mb_ptr2len)(STR_PTR(i));
 	}
 	else
 #endif
 	{
-	    *p = TO_LOWER(*p);
-	    ++p;
+	    if (buf == NULL)
+		GA_CHAR(i) = TOLOWER_LOC(GA_CHAR(i));
+	    else
+		buf[i] = TOLOWER_LOC(buf[i]);
+	    ++i;
 	}
     }
+
+    if (buf == NULL)
+	return (char_u *)ga.ga_data;
+    return buf;
 }
 #endif
 
@@ -493,7 +586,7 @@ transchar_nonprint(buf, c)
 {
     if (c == NL)
 	c = NUL;		/* we use newline in place of a NUL */
-    else if (c == CR && get_fileformat(curbuf) == EOL_MAC)
+    else if (c == CAR && get_fileformat(curbuf) == EOL_MAC)
 	c = NL;			/* we use CR in place of  NL in this case */
 
     if (dy_flags & DY_UHEX)		/* 'display' has "uhex" */
@@ -657,28 +750,43 @@ ptr2cells(p)
 }
 
 /*
- * Return the number of characters string 's' will take on the screen,
+ * Return the number of characters string "s" will take on the screen,
  * counting TABs as two characters: "^I".
  */
     int
 vim_strsize(s)
     char_u	*s;
 {
-    int		len = 0;
+    return vim_strnsize(s, (int)MAXCOL);
+}
 
-    while (*s != NUL)
+/*
+ * Return the number of characters string "s[len]" will take on the screen,
+ * counting TABs as two characters: "^I".
+ */
+    int
+vim_strnsize(s, len)
+    char_u	*s;
+    int		len;
+{
+    int		size = 0;
+
+    while (*s != NUL && --len >= 0)
     {
 #ifdef FEAT_MBYTE
 	if (has_mbyte)
 	{
-	    len += ptr2cells(s);
-	    s += (*mb_ptr2len_check)(s);
+	    int	    l = (*mb_ptr2len)(s);
+
+	    size += ptr2cells(s);
+	    s += l;
+	    len -= l - 1;
 	}
 	else
 #endif
-	    len += byte2cells(*s++);
+	    size += byte2cells(*s++);
     }
-    return len;
+    return size;
 }
 
 /*
@@ -747,22 +855,14 @@ win_linetabsize(wp, p, len)
     colnr_T	col = 0;
     char_u	*s;
 
-    for (s = p; *s != NUL && s < p + len; )
-    {
+    for (s = p; *s != NUL && (len == MAXCOL || s < p + len); mb_ptr_adv(s))
 	col += win_lbr_chartabsize(wp, s, col, NULL);
-#ifdef FEAT_MBYTE
-	if (has_mbyte)
-	    s += (*mb_ptr2len_check)(s);
-	else
-#endif
-	    ++s;
-    }
     return (int)col;
 }
 
 /*
- * return TRUE if 'c' is a normal identifier character
- * letters and characters from 'isident' option.
+ * Return TRUE if 'c' is a normal identifier character:
+ * Letters and characters from the 'isident' option.
  */
     int
 vim_isIDc(c)
@@ -831,6 +931,23 @@ vim_isfilec(c)
     return (c >= 0x100 || (c > 0 && (chartab[c] & CT_FNAME_CHAR)));
 }
 
+/*
+ * return TRUE if 'c' is a valid file-name character or a wildcard character
+ * Assume characters above 0x100 are valid (multi-byte).
+ * Explicitly interpret ']' as a wildcard character as mch_has_wildcard("]")
+ * returns false.
+ */
+    int
+vim_isfilec_or_wc(c)
+    int c;
+{
+    char_u buf[2];
+
+    buf[0] = (char_u)c;
+    buf[1] = NUL;
+    return vim_isfilec(c) || c == ']' || mch_has_wildcard(buf);
+}
+
 /*
  * return TRUE if 'c' is a printable character
  * Assume characters above 0x100 are printable (multi-byte), except for
@@ -898,12 +1015,7 @@ lbr_chartabsize_adv(s, col)
     int		retval;
 
     retval = lbr_chartabsize(*s, col);
-#ifdef FEAT_MBYTE
-    if (has_mbyte)
-	*s += (*mb_ptr2len_check)(*s);
-    else
-#endif
-	++*s;
+    mb_ptr_adv(*s);
     return retval;
 }
 
@@ -936,6 +1048,7 @@ win_lbr_chartabsize(wp, s, col, headp)
     int		numberextra;
     char_u	*ps;
     int		tab_corr = (*s == TAB);
+    int		n;
 
     /*
      * No 'linebreak' and 'showbreak': return quickly.
@@ -979,18 +1092,16 @@ win_lbr_chartabsize(wp, s, col, headp)
 	col2 = col;
 	colmax = W_WIDTH(wp) - numberextra;
 	if (col >= colmax)
-	    colmax += (((col - colmax)
-			/ (colmax + win_col_off2(wp))) + 1)
-			* (colmax + win_col_off2(wp));
+	{
+	    n = colmax + win_col_off2(wp);
+	    if (n > 0)
+		colmax += (((col - colmax) / n) + 1) * n;
+	}
+
 	for (;;)
 	{
 	    ps = s;
-# ifdef FEAT_MBYTE
-	    if (has_mbyte)
-		s += (*mb_ptr2len_check)(s);
-	    else
-# endif
-		++s;
+	    mb_ptr_adv(s);
 	    c = *s;
 	    if (!(c != NUL
 		    && (vim_isbreak(c)
@@ -1034,7 +1145,7 @@ win_lbr_chartabsize(wp, s, col, headp)
 	}
 	if (col == 0 || col + size > (colnr_T)W_WIDTH(wp))
 	{
-	    added = (int)STRLEN(p_sbr);
+	    added = vim_strsize(p_sbr);
 	    if (tab_corr)
 		size += (added / wp->w_buffer->b_p_ts) * wp->w_buffer->b_p_ts;
 	    else
@@ -1179,7 +1290,8 @@ getvcol(wp, pos, start, cursor, end)
 		    /* If a double-cell char doesn't fit at the end of a line
 		     * it wraps to the next line, it's like this char is three
 		     * cells wide. */
-		    if (incr == 2 && wp->w_p_wrap && in_win_border(wp, vcol))
+		    if (incr == 2 && wp->w_p_wrap && MB_BYTE2LEN(*ptr) > 1
+			    && in_win_border(wp, vcol))
 		    {
 			++incr;
 			head = 1;
@@ -1194,12 +1306,7 @@ getvcol(wp, pos, start, cursor, end)
 		break;
 
 	    vcol += incr;
-#ifdef FEAT_MBYTE
-	    if (has_mbyte)
-		ptr += (*mb_ptr2len_check)(ptr);
-	    else
-#endif
-		++ptr;
+	    mb_ptr_adv(ptr);
 	}
     }
     else
@@ -1220,12 +1327,7 @@ getvcol(wp, pos, start, cursor, end)
 		break;
 
 	    vcol += incr;
-#ifdef FEAT_MBYTE
-	    if (has_mbyte)
-		ptr += (*mb_ptr2len_check)(ptr);
-	    else
-#endif
-		++ptr;
+	    mb_ptr_adv(ptr);
 	}
     }
     if (start != NULL)
@@ -1239,7 +1341,8 @@ getvcol(wp, pos, start, cursor, end)
 		&& !wp->w_p_list
 		&& !virtual_active()
 #ifdef FEAT_VISUAL
-		&& !(VIsual_active && *p_sel == 'e')
+		&& !(VIsual_active
+				   && (*p_sel == 'e' || ltoreq(*pos, VIsual)))
 #endif
 		)
 	    *cursor = vcol + incr - 1;	    /* cursor at end */
@@ -1277,31 +1380,43 @@ getvvcol(wp, pos, start, cursor, end)
     colnr_T	*end;
 {
     colnr_T	col;
+    colnr_T	coladd;
+    colnr_T	endadd;
+# ifdef FEAT_MBYTE
     char_u	*ptr;
+# endif
 
     if (virtual_active())
     {
 	/* For virtual mode, only want one value */
 	getvcol(wp, pos, &col, NULL, NULL);
 
-	if (pos->coladd > 0)
+	coladd = pos->coladd;
+	endadd = 0;
+# ifdef FEAT_MBYTE
+	/* Cannot put the cursor on part of a wide character. */
+	ptr = ml_get_buf(wp->w_buffer, pos->lnum, FALSE);
+	if (pos->col < STRLEN(ptr))
 	{
-	    /* Adjust for multiwide char */
-	    ptr = ml_get(pos->lnum);
-	    if (pos->col <= STRLEN(ptr))
+	    int c = (*mb_ptr2char)(ptr + pos->col);
+
+	    if (c != TAB && vim_isprintc(c))
 	    {
-		ptr += pos->col;
-		if (*ptr != TAB && *ptr != NUL && ptr2cells(ptr) > 1)
-		    pos->coladd = 0;
+		endadd = char2cells(c) - 1;
+		if (coladd > endadd)	/* past end of line */
+		    endadd = 0;
+		else
+		    coladd = 0;
 	    }
-	    col += pos->coladd;
 	}
+# endif
+	col += coladd;
 	if (start != NULL)
 	    *start = col;
 	if (cursor != NULL)
 	    *cursor = col;
 	if (end != NULL)
-	    *end = col;
+	    *end = col + endadd;
     }
     else
 	getvcol(wp, pos, start, cursor, end);
@@ -1360,27 +1475,200 @@ skipwhite(p)
 }
 
 /*
- * skipdigits: skip over digits;
+ * skip over digits
  */
     char_u *
 skipdigits(p)
     char_u	*p;
 {
-    while (isdigit(*p))	/* skip to next non-digit */
+    while (VIM_ISDIGIT(*p))	/* skip to next non-digit */
+	++p;
+    return p;
+}
+
+#if defined(FEAT_SYN_HL) || defined(FEAT_SPELL) || defined(PROTO)
+/*
+ * skip over digits and hex characters
+ */
+    char_u *
+skiphex(p)
+    char_u	*p;
+{
+    while (vim_isxdigit(*p))	/* skip to next non-digit */
+	++p;
+    return p;
+}
+#endif
+
+#if defined(FEAT_EX_EXTRA) || defined(PROTO)
+/*
+ * skip to digit (or NUL after the string)
+ */
+    char_u *
+skiptodigit(p)
+    char_u	*p;
+{
+    while (*p != NUL && !VIM_ISDIGIT(*p))	/* skip to next digit */
 	++p;
     return p;
 }
 
 /*
- * vim_isdigit: version of isdigit() that can handle characters > 0x100.
+ * skip to hex character (or NUL after the string)
+ */
+    char_u *
+skiptohex(p)
+    char_u	*p;
+{
+    while (*p != NUL && !vim_isxdigit(*p))	/* skip to next digit */
+	++p;
+    return p;
+}
+#endif
+
+/*
+ * Variant of isdigit() that can handle characters > 0x100.
+ * We don't use isdigit() here, because on some systems it also considers
+ * superscript 1 to be a digit.
+ * Use the VIM_ISDIGIT() macro for simple arguments.
  */
     int
 vim_isdigit(c)
+    int		c;
+{
+    return (c >= '0' && c <= '9');
+}
+
+/*
+ * Variant of isxdigit() that can handle characters > 0x100.
+ * We don't use isxdigit() here, because on some systems it also considers
+ * superscript 1 to be a digit.
+ */
+    int
+vim_isxdigit(c)
+    int		c;
+{
+    return (c >= '0' && c <= '9')
+	|| (c >= 'a' && c <= 'f')
+	|| (c >= 'A' && c <= 'F');
+}
+
+#if defined(FEAT_MBYTE) || defined(PROTO)
+/*
+ * Vim's own character class functions.  These exist because many library
+ * islower()/toupper() etc. do not work properly: they crash when used with
+ * invalid values or can't handle latin1 when the locale is C.
+ * Speed is most important here.
+ */
+#define LATIN1LOWER 'l'
+#define LATIN1UPPER 'U'
+
+/*                                                                 !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]%_'abcdefghijklmnopqrstuvwxyz{|}~                                   */
+static char_u latin1flags[257] = "                                                                 UUUUUUUUUUUUUUUUUUUUUUUUUU      llllllllllllllllllllllllll                                                                     UUUUUUUUUUUUUUUUUUUUUUU UUUUUUUllllllllllllllllllllllll llllllll";
+static char_u latin1upper[257] = "                                 !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`ABCDEFGHIJKLMNOPQRSTUVWXYZ{|}~";
+static char_u latin1lower[257] = "                                 !\"#$%&'()*+,-./0123456789:;<=>?@abcdefghijklmnopqrstuvwxyz[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
+
+    int
+vim_islower(c)
     int	    c;
 {
-    return (c > 0 && c < 0x100 && isdigit(c));
+    if (c <= '@')
+	return FALSE;
+    if (c >= 0x80)
+    {
+	if (enc_utf8)
+	    return utf_islower(c);
+	if (c >= 0x100)
+	{
+#ifdef HAVE_ISWLOWER
+	    if (has_mbyte)
+		return iswlower(c);
+#endif
+	    /* islower() can't handle these chars and may crash */
+	    return FALSE;
+	}
+	if (enc_latin1like)
+	    return (latin1flags[c] & LATIN1LOWER) == LATIN1LOWER;
+    }
+    return islower(c);
 }
 
+    int
+vim_isupper(c)
+    int	    c;
+{
+    if (c <= '@')
+	return FALSE;
+    if (c >= 0x80)
+    {
+	if (enc_utf8)
+	    return utf_isupper(c);
+	if (c >= 0x100)
+	{
+#ifdef HAVE_ISWUPPER
+	    if (has_mbyte)
+		return iswupper(c);
+#endif
+	    /* islower() can't handle these chars and may crash */
+	    return FALSE;
+	}
+	if (enc_latin1like)
+	    return (latin1flags[c] & LATIN1UPPER) == LATIN1UPPER;
+    }
+    return isupper(c);
+}
+
+    int
+vim_toupper(c)
+    int	    c;
+{
+    if (c <= '@')
+	return c;
+    if (c >= 0x80)
+    {
+	if (enc_utf8)
+	    return utf_toupper(c);
+	if (c >= 0x100)
+	{
+#ifdef HAVE_TOWUPPER
+	    if (has_mbyte)
+		return towupper(c);
+#endif
+	    /* toupper() can't handle these chars and may crash */
+	    return c;
+	}
+	if (enc_latin1like)
+	    return latin1upper[c];
+    }
+    return TOUPPER_LOC(c);
+}
+
+    int
+vim_tolower(c)
+    int	    c;
+{
+    if (c <= '@')
+	return c;
+    if (c >= 0x80)
+    {
+	if (enc_utf8)
+	    return utf_tolower(c);
+	if (c >= 0x100)
+	{
+#ifdef HAVE_TOWLOWER
+	    if (has_mbyte)
+		return towlower(c);
+#endif
+	    /* tolower() can't handle these chars and may crash */
+	    return c;
+	}
+	if (enc_latin1like)
+	    return latin1lower[c];
+    }
+    return TOLOWER_LOC(c);
+}
+#endif
+
 /*
  * skiptowhite: skip over text until ' ' or '\t' or NUL.
  */
@@ -1393,7 +1681,8 @@ skiptowhite(p)
     return p;
 }
 
-#if defined(FEAT_LISTCMDS) || defined(PROTO)
+#if defined(FEAT_LISTCMDS) || defined(FEAT_SIGNS) || defined(FEAT_SNIFF) \
+	|| defined(PROTO)
 /*
  * skiptowhite_esc: Like skiptowhite(), but also skip escaped chars
  */
@@ -1446,7 +1735,7 @@ vim_isblankline(lbuf)
 
 /*
  * Convert a string into a long and/or unsigned long, taking care of
- * hexadecimal and octal numbers.
+ * hexadecimal and octal numbers.  Accepts a '-' sign.
  * If "hexp" is not NULL, returns a flag to indicate the type of the number:
  *  0	    decimal
  *  '0'	    octal
@@ -1455,6 +1744,11 @@ vim_isblankline(lbuf)
  * If "len" is not NULL, the length of the number in characters is returned.
  * If "nptr" is not NULL, the signed result is returned in it.
  * If "unptr" is not NULL, the unsigned result is returned in it.
+ * If "unptr" is not NULL, the unsigned result is returned in it.
+ * If "dooct" is non-zero recognize octal numbers, when > 1 always assume
+ * octal number.
+ * If "dohex" is non-zero recognize hex numbers, when > 1 always assume
+ * hex number.
  */
     void
 vim_str2nr(start, hexp, len, dooct, dohex, nptr, unptr)
@@ -1470,8 +1764,8 @@ vim_str2nr(start, hexp, len, dooct, dohex, nptr, unptr)
     char_u	    *ptr = start;
     int		    hex = 0;		/* default is decimal */
     int		    negative = FALSE;
-    long	    n = 0;
     unsigned long   un = 0;
+    int		    n;
 
     if (ptr[0] == '-')
     {
@@ -1479,66 +1773,74 @@ vim_str2nr(start, hexp, len, dooct, dohex, nptr, unptr)
 	++ptr;
     }
 
-    if (ptr[0] == '0')			/* could be hex or octal */
+    /* Recognize hex and octal. */
+    if (ptr[0] == '0' && ptr[1] != '8' && ptr[1] != '9')
     {
 	hex = ptr[1];
-	if (dohex && (hex == 'X' || hex == 'x') && isxdigit(ptr[2]))
+	if (dohex && (hex == 'X' || hex == 'x') && vim_isxdigit(ptr[2]))
 	    ptr += 2;			/* hexadecimal */
 	else
 	{
-	    if (dooct && isdigit(hex))
-		hex = '0';		/* octal */
-	    else
-		hex = 0;		/* 0 by itself is decimal */
+	    hex = 0;			/* default is decimal */
+	    if (dooct)
+	    {
+		/* Don't interpret "0", "08" or "0129" as octal. */
+		for (n = 1; VIM_ISDIGIT(ptr[n]); ++n)
+		{
+		    if (ptr[n] > '7')
+		    {
+			hex = 0;	/* can't be octal */
+			break;
+		    }
+		    if (ptr[n] > '0')
+			hex = '0';	/* assume octal */
+		}
+	    }
 	}
     }
 
     /*
      * Do the string-to-numeric conversion "manually" to avoid sscanf quirks.
      */
-    if (hex)
+    if (hex == '0' || dooct > 1)
     {
-	if (hex == '0')
+	/* octal */
+	while ('0' <= *ptr && *ptr <= '7')
 	{
-	    /* octal */
-	    while ('0' <= *ptr && *ptr <= '7')
-	    {
-		n = 8 * n + (long)(*ptr - '0');
-		un = 8 * un + (unsigned long)(*ptr - '0');
-		++ptr;
-	    }
+	    un = 8 * un + (unsigned long)(*ptr - '0');
+	    ++ptr;
 	}
-	else
+    }
+    else if (hex != 0 || dohex > 1)
+    {
+	/* hex */
+	while (vim_isxdigit(*ptr))
 	{
-	    /* hex */
-	    while (isxdigit(*ptr))
-	    {
-		n = 16 * n + (long)hex2nr(*ptr);
-		un = 16 * un + (unsigned long)hex2nr(*ptr);
-		++ptr;
-	    }
+	    un = 16 * un + (unsigned long)hex2nr(*ptr);
+	    ++ptr;
 	}
     }
     else
     {
 	/* decimal */
-	while (isdigit(*ptr))
+	while (VIM_ISDIGIT(*ptr))
 	{
-	    n = 10 * n + (long)(*ptr - '0');
 	    un = 10 * un + (unsigned long)(*ptr - '0');
 	    ++ptr;
 	}
     }
 
-    if (!hex && negative)   /* account for leading '-' for decimal numbers */
-	n = -n;
-
     if (hexp != NULL)
 	*hexp = hex;
     if (len != NULL)
 	*len = (int)(ptr - start);
     if (nptr != NULL)
-	*nptr = n;
+    {
+	if (negative)   /* account for leading '-' for decimal numbers */
+	    *nptr = -(long)un;
+	else
+	    *nptr = (long)un;
+    }
     if (unptr != NULL)
 	*unptr = un;
 }
@@ -1558,7 +1860,8 @@ hex2nr(c)
     return c - '0';
 }
 
-#if defined(FEAT_TERMRESPONSE) || defined(PROTO)
+#if defined(FEAT_TERMRESPONSE) \
+	|| (defined(FEAT_GUI_GTK) && defined(FEAT_WINDOWS)) || defined(PROTO)
 /*
  * Convert two hex characters to a byte.
  * Return -1 if one of the characters is not hex.
@@ -1567,7 +1870,7 @@ hex2nr(c)
 hexhex2nr(p)
     char_u	*p;
 {
-    if (!isxdigit(p[0]) || !isxdigit(p[1]))
+    if (!vim_isxdigit(p[0]) || !vim_isxdigit(p[1]))
 	return -1;
     return (hex2nr(p[0]) << 4) + hex2nr(p[1]);
 }
@@ -1616,7 +1919,7 @@ backslash_halve(p)
 {
     for ( ; *p; ++p)
 	if (rem_backslash(p))
-	    STRCPY(p, p + 1);
+	    mch_memmove(p, p + 1, STRLEN(p));
 }
 
 /*
