@@ -106,11 +106,17 @@ static tcl_info tclinfo = { NULL, 0, 0, 0, NULL, NULL };
 #define VIMOUT	((ClientData)1)
 #define VIMERR	((ClientData)2)
 
+/* This appears to be new in Tcl 8.4. */
+#ifndef CONST84
+# define CONST84
+#endif
+
 /*
  *  List of Tcl interpreters who reference a vim window or buffer.
- *  Each buffer and window has it's own list in the tcl_ref struct member.
- *  We need this because Tcl can create sub-interpreters with the "interp"
- *  command, and each interpreter can reference all windows and buffers.
+ *  Each buffer and window has it's own list in the w_tcl_ref or b_tcl_ref
+ *  struct member.  We need this because Tcl can create sub-interpreters with
+ *  the "interp" command, and each interpreter can reference all windows and
+ *  buffers.
  */
 struct ref
 {
@@ -151,7 +157,7 @@ static struct ref refsdeleted;	/* dummy object for deleted ref list */
 # endif
 
 # ifndef  DYNAMIC_TCL /* Just generating prototypes */
-#  define HANDLE int
+typedef int HANDLE;
 # endif
 
 /*
@@ -192,7 +198,7 @@ tcl_runtime_link_init(char *libname, int verbose)
     if (!(hTclLib = LoadLibraryEx(libname, NULL, 0)))
     {
 	if (verbose)
-	    EMSG2(_("E370: Could not load library %s"), libname);
+	    EMSG2(_(e_loadlib), libname);
 	return FAIL;
     }
     for (i = 0; tcl_funcname_table[i].ptr; ++i)
@@ -203,8 +209,7 @@ tcl_runtime_link_init(char *libname, int verbose)
 	    FreeLibrary(hTclLib);
 	    hTclLib = NULL;
 	    if (verbose)
-		EMSG2(_("E448: Could not load library function %s"),
-						  tcl_funcname_table[i].name);
+		EMSG2(_(e_loadfunc), tcl_funcname_table[i].name);
 	    return FAIL;
 	}
     }
@@ -217,7 +222,7 @@ static char *find_executable_arg = NULL;
 #endif
 
     void
-tcl_init(arg)
+vim_tcl_init(arg)
     char	*arg;
 {
 #ifndef DYNAMIC_TCL
@@ -370,7 +375,7 @@ beepcmd(dummy, interp, objc, objv)
 
 /*
  *  "::vim::buffer list" - create a list of buffer commands.
- *  "::vim::bufffer {N}" - create buffer command for buffer N.
+ *  "::vim::buffer {N}" - create buffer command for buffer N.
  *  "::vim::buffer new" - create a new buffer (not implemented)
  */
 /* ARGSUSED */
@@ -385,9 +390,8 @@ buffercmd(dummy, interp, objc, objv)
     buf_T	*buf;
     Tcl_Obj	*resobj;
     int		err, n, idx;
-
     enum {BCMD_EXISTS, BCMD_LIST};
-    static char *bcmdoptions[] =
+    static CONST84 char *bcmdoptions[] =
     {
 	"exists", "list", (char *)0
     };
@@ -534,7 +538,7 @@ bufselfcmd(ref, interp, objc, objv)
 	BUF_GET, BUF_INSERT, BUF_LAST, BUF_MARK, BUF_NAME, BUF_NUMBER,
 	BUF_OPTION, BUF_SET, BUF_WINDOWS
     };
-    static char *bufoptions[] =
+    static CONST84 char *bufoptions[] =
     {
 	"append", "command", "count", "delcmd", "delete", "expr",
 	"get", "insert", "last", "mark", "name", "number",
@@ -849,7 +853,9 @@ bufselfcmd(ref, interp, objc, objv)
 	    else
 	    {
 		char rbuf[64];
-		sprintf(rbuf, _("row %d column %d"), (int)row2tcl(pos->lnum), (int)col2tcl(pos->col));
+
+		sprintf(rbuf, _("row %d column %d"),
+			     (int)row2tcl(pos->lnum), (int)col2tcl(pos->col));
 		Tcl_SetResult(interp, rbuf, TCL_VOLATILE);
 	    }
 	    break;
@@ -871,7 +877,8 @@ bufselfcmd(ref, interp, objc, objv)
 		--val1;
 	    if (u_save((linenr_T)val1, (linenr_T)(val1+1)) != OK)
 	    {
-		Tcl_SetResult(interp, _("cannot save undo information"), TCL_STATIC);
+		Tcl_SetResult(interp, _("cannot save undo information"),
+								  TCL_STATIC);
 		err = TCL_ERROR;
 		break;
 	    }
@@ -879,7 +886,8 @@ bufselfcmd(ref, interp, objc, objv)
 	    line = Tcl_GetStringFromObj(objv[3], NULL);
 	    if (ml_append((linenr_T)val1, (char_u *)line, 0, FALSE) != OK)
 	    {
-		Tcl_SetResult(interp, _("cannot insert/append line"), TCL_STATIC);
+		Tcl_SetResult(interp, _("cannot insert/append line"),
+								  TCL_STATIC);
 		err = TCL_ERROR;
 		break;
 	    }
@@ -925,7 +933,7 @@ bufselfcmd(ref, interp, objc, objv)
 		err = TCL_ERROR;
 		break;
 	    }
-	    err = tclsetdelcmd(interp, buf->tcl_ref, (void *)buf, objv[2]);
+	    err = tclsetdelcmd(interp, buf->b_tcl_ref, (void *)buf, objv[2]);
 	    break;
 
 	default:
@@ -967,7 +975,7 @@ winselfcmd(ref, interp, objc, objv)
 	WIN_BUFFER, WIN_COMMAND, WIN_CURSOR, WIN_DELCMD, WIN_EXPR,
 	WIN_HEIGHT, WIN_OPTION
     };
-    static char *winoptions[] =
+    static CONST84 char *winoptions[] =
     {
 	"buffer", "command", "cursor", "delcmd", "expr",
 	"height", "option", (char *)0
@@ -1051,7 +1059,7 @@ winselfcmd(ref, interp, objc, objv)
 		err = TCL_ERROR;
 		break;
 	    }
-	    err = tclsetdelcmd(interp, win->tcl_ref, (void *)win, objv[2]);
+	    err = tclsetdelcmd(interp, win->w_tcl_ref, (void *)win, objv[2]);
 	    break;
 
 	case WIN_CURSOR:
@@ -1064,6 +1072,7 @@ winselfcmd(ref, interp, objc, objv)
 	    if (objc == 2)
 	    {
 		char buf[64];
+
 		sprintf(buf, _("row %d column %d"), (int)row2tcl(win->w_cursor.lnum), (int)col2tcl(win->w_cursor.col));
 		Tcl_SetResult(interp, buf, TCL_VOLATILE);
 		break;
@@ -1180,7 +1189,7 @@ tclgetlinenum(interp, obj, valueP, buf)
 
     enum { LN_BEGIN, LN_BOTTOM, LN_END, LN_FIRST, LN_LAST, LN_START, LN_TOP };
 
-    static char *keyw[] =
+    static CONST84 char *keyw[] =
     {
 	"begin", "bottom", "end", "first", "last", "start", "top", (char *)0
     };
@@ -1312,7 +1321,7 @@ tclsetoption(interp, objc, objv, objn)
     Tcl_Obj	*resobj;
 
     enum { OPT_OFF, OPT_ON, OPT_TOGGLE };
-    static char *optkw[] = { "off", "on", "toggle", (char *)0 };
+    static CONST84 char *optkw[] = { "off", "on", "toggle", (char *)0 };
 
     nobjs = objc - objn;
     if (nobjs != 1 && nobjs != 2)
@@ -1398,7 +1407,7 @@ tclvimexpr(interp, objc, objv, objn)
 
 #ifdef FEAT_EVAL
     expr = Tcl_GetStringFromObj(objv[objn], NULL);
-    str = (char *)eval_to_string((char_u *)expr, NULL);
+    str = (char *)eval_to_string((char_u *)expr, NULL, TRUE);
     if (str == NULL)
 	Tcl_SetResult(interp, _("invalid expression"), TCL_STATIC);
     else
@@ -1457,7 +1466,8 @@ delref(cref)
     static char *
 tclgetref(interp, refstartP, prefix, vimobj, proc)
     Tcl_Interp	*interp;
-    void	**refstartP;	/* ptr to tcl_ref member of win_T/buf_T struct */
+    void	**refstartP;	/* ptr to w_tcl_ref/b_tcl-ref member of
+				   win_T/buf_T struct */
     char	*prefix;	/* "win" or "buf" */
     void	*vimobj;	/* win_T* or buf_T* */
     Tcl_ObjCmdProc *proc;	/* winselfcmd or bufselfcmd */
@@ -1483,7 +1493,8 @@ tclgetref(interp, refstartP, prefix, vimobj, proc)
     }
 
     if (ref)
-	sprintf(name, "::vim::%s", Tcl_GetCommandName(interp, ref->cmd));
+	vim_snprintf(name, sizeof(name), "::vim::%s",
+					Tcl_GetCommandName(interp, ref->cmd));
     else
     {
 	if (unused)
@@ -1504,7 +1515,8 @@ tclgetref(interp, refstartP, prefix, vimobj, proc)
 	}
 
 	/* This might break on some exotic systems... */
-	sprintf(name, "::vim::%s_%lx", prefix, (unsigned long)vimobj);
+	vim_snprintf(name, sizeof(name), "::vim::%s_%lx",
+					       prefix, (unsigned long)vimobj);
 	cmd = Tcl_CreateObjCommand(interp, name, proc,
 	    (ClientData)ref, (Tcl_CmdDeleteProc *)delref);
 	if (!cmd)
@@ -1523,7 +1535,7 @@ tclgetwindow(interp, win)
     Tcl_Interp	*interp;
     win_T	*win;
 {
-    return tclgetref(interp, &(win->tcl_ref), "win", (void *)win, winselfcmd);
+    return tclgetref(interp, &(win->w_tcl_ref), "win", (void *)win, winselfcmd);
 }
 
     static char *
@@ -1531,7 +1543,7 @@ tclgetbuffer(interp, buf)
     Tcl_Interp	*interp;
     buf_T	*buf;
 {
-    return tclgetref(interp, &(buf->tcl_ref), "buf", (void *)buf, bufselfcmd);
+    return tclgetref(interp, &(buf->b_tcl_ref), "buf", (void *)buf, bufselfcmd);
 }
 
     static int
@@ -1552,7 +1564,9 @@ tclsetdelcmd(interp, reflist, vimobj, delcmd)
 	if (reflist->interp == interp && reflist->vimobj == vimobj)
 	{
 	    if (reflist->delcmd)
+	    {
 		Tcl_DecrRefCount(reflist->delcmd);
+	    }
 	    Tcl_IncrRefCount(delcmd);
 	    reflist->delcmd = delcmd;
 	    return TCL_OK;
@@ -1722,7 +1736,7 @@ tclinit(eap)
 #ifdef DYNAMIC_TCL
     if (!tcl_enabled(TRUE))
     {
-	EMSG(_("Sorry, this command is disabled: the Tcl library could not be loaded."));
+	EMSG(_("E571: Sorry, this command is disabled: the Tcl library could not be loaded."));
 	return FAIL;
     }
 #endif
@@ -1858,7 +1872,7 @@ tcldelthisinterp()
      * callback, which deletes all refs pointing to this interpreter.
      * We could garbage-collect the unused ref structs in all windows and
      * buffers, but unless the user creates hundreds of sub-interpreters
-     * all refering to lots of windows and buffers, this is hardly worth
+     * all referring to lots of windows and buffers, this is hardly worth
      * the effort.  Unused refs are recycled by other interpreters, and
      * all refs are free'd when the window/buffer gets closed by vim.
      */
@@ -1878,7 +1892,7 @@ tclexit(error)
     if (error == TCL_EXIT )
     {
 	int retval;
-	char buf[32];
+	char buf[50];
 	Tcl_Obj *robj;
 
 	robj = Tcl_GetObjResult(tclinfo.interp);
@@ -1889,7 +1903,7 @@ tclexit(error)
 	}
 	else
 	{
-	    sprintf(buf, "exit code %d", retval);
+	    sprintf(buf, _("E572: exit code %d"), retval);
 	    tclerrmsg(buf);
 	    if (retval == 0 )
 	    {
@@ -1906,7 +1920,7 @@ tclexit(error)
     {
 	char *result;
 
-	result = Tcl_GetStringResult(tclinfo.interp);
+	result = (char *)Tcl_GetStringResult(tclinfo.interp);
 	if (error == TCL_OK)
 	{
 	    tclmsg(result);
@@ -1932,20 +1946,21 @@ ex_tcl(eap)
     char_u	*script;
     int		err;
 
-    err = tclinit(eap);
-    if (err == OK)
+    script = script_get(eap, eap->arg);
+    if (!eap->skip)
     {
-	Tcl_AllowExceptions(tclinfo.interp);
-	script = script_get(eap, eap->arg);
-	if (script == NULL)
-	    err = Tcl_Eval(tclinfo.interp, (char *)eap->arg);
-	else
+	err = tclinit(eap);
+	if (err == OK)
 	{
-	    err = Tcl_Eval(tclinfo.interp, (char *)script);
-	    vim_free(script);
+	    Tcl_AllowExceptions(tclinfo.interp);
+	    if (script == NULL)
+		err = Tcl_Eval(tclinfo.interp, (char *)eap->arg);
+	    else
+		err = Tcl_Eval(tclinfo.interp, (char *)script);
+	    err = tclexit(err);
 	}
-	err = tclexit(err);
     }
+    vim_free(script);
 }
 
 /*
@@ -2013,7 +2028,7 @@ ex_tcldo(eap)
 	err = Tcl_Eval(tclinfo.interp, script);
 	if (err != TCL_OK)
 	    break;
-	line = Tcl_GetVar(tclinfo.interp, var_line, 0);
+	line = (char *)Tcl_GetVar(tclinfo.interp, var_line, 0);
 	if (line)
 	{
 	    if (ml_replace((linenr_T)rs, (char_u *)line, TRUE) != OK)
@@ -2059,7 +2074,7 @@ tcldelallrefs(ref)
 		err = Tcl_GlobalEvalObj(ref->interp, ref->delcmd);
 		if (err != TCL_OK)
 		{
-		    result = Tcl_GetStringResult(ref->interp);
+		    result = (char *)Tcl_GetStringResult(ref->interp);
 		    if (result)
 			tclerrmsg(result);
 		}
@@ -2084,12 +2099,12 @@ tcl_buffer_free(buf)
 	return;
 #endif
 
-    reflist = (struct ref*)(buf->tcl_ref);
+    reflist = (struct ref *)(buf->b_tcl_ref);
     if (reflist != &refsdeleted)
     {
-	buf->tcl_ref = (void *)&refsdeleted;
+	buf->b_tcl_ref = (void *)&refsdeleted;
 	tcldelallrefs(reflist);
-	buf->tcl_ref = NULL;
+	buf->b_tcl_ref = NULL;
     }
 }
 
@@ -2105,12 +2120,12 @@ tcl_window_free(win)
 	return;
 #endif
 
-    reflist = (struct ref*)(win->tcl_ref);
+    reflist = (struct ref*)(win->w_tcl_ref);
     if (reflist != &refsdeleted)
     {
-	win->tcl_ref = (void *)&refsdeleted;
+	win->w_tcl_ref = (void *)&refsdeleted;
 	tcldelallrefs(reflist);
-	win->tcl_ref = NULL;
+	win->w_tcl_ref = NULL;
     }
 }
 #endif
