@@ -4,6 +4,7 @@
  *
  * Do ":help uganda"  in Vim to read copying and usage conditions.
  * Do ":help credits" in Vim to see a list of people who contributed.
+ * See README.txt for an overview of the Vim source code.
  */
 
 /*
@@ -31,8 +32,8 @@
  * file is opened.
  */
 
-#if defined MSDOS || defined WIN32
-# include <io.h>	/* for lseek(), must be before vim.h */
+#if defined MSDOS || defined(WIN32) || defined(_WIN64)
+# include "vimio.h"	/* for lseek(), must be before vim.h */
 #endif
 
 #include "vim.h"
@@ -47,14 +48,14 @@
 #ifdef HAVE_ST_BLKSIZE
 # define STATFS stat
 # define F_BSIZE st_blksize
-# define fstatfs(fd, buf, len, nul) fstat((fd), (buf))
+# define fstatfs(fd, buf, len, nul) mch_fstat((fd), (buf))
 #else
 # ifdef HAVE_SYS_STATFS_H
 #  include <sys/statfs.h>
 #  define STATFS statfs
 #  define F_BSIZE f_bsize
 #  ifdef __MINT__		/* do we still need this? */
-#   define fstatfs(fd, buf, len, nul) fstat((fd), (buf))
+#   define fstatfs(fd, buf, len, nul) mch_fstat((fd), (buf))
 #  endif
 # endif
 #endif
@@ -63,7 +64,7 @@
  * for Amiga Dos 2.0x we use Flush
  */
 #ifdef AMIGA
-# ifndef NO_ARP
+# ifdef FEAT_ARP
 extern int dos2;			/* this is in os_amiga.c */
 # endif
 # ifdef SASC
@@ -75,22 +76,21 @@ extern int dos2;			/* this is in os_amiga.c */
 #define MEMFILE_PAGE_SIZE 4096		/* default page size */
 
 static long_u	total_mem_used = 0;	/* total memory used for memfiles */
-static int	dont_release = FALSE;	/* don't release blocks */
 
-static void mf_ins_hash __ARGS((MEMFILE *, BHDR *));
-static void mf_rem_hash __ARGS((MEMFILE *, BHDR *));
-static BHDR *mf_find_hash __ARGS((MEMFILE *, blocknr_t));
-static void mf_ins_used __ARGS((MEMFILE *, BHDR *));
-static void mf_rem_used __ARGS((MEMFILE *, BHDR *));
-static BHDR *mf_release __ARGS((MEMFILE *, int));
-static BHDR *mf_alloc_bhdr __ARGS((MEMFILE *, int));
-static void mf_free_bhdr __ARGS((BHDR *));
-static void mf_ins_free __ARGS((MEMFILE *, BHDR *));
-static BHDR *mf_rem_free __ARGS((MEMFILE *));
-static int  mf_read __ARGS((MEMFILE *, BHDR *));
-static int  mf_write __ARGS((MEMFILE *, BHDR *));
-static int  mf_trans_add __ARGS((MEMFILE *, BHDR *));
-static void mf_do_open __ARGS((MEMFILE *, char_u *, int));
+static void mf_ins_hash __ARGS((memfile_T *, bhdr_T *));
+static void mf_rem_hash __ARGS((memfile_T *, bhdr_T *));
+static bhdr_T *mf_find_hash __ARGS((memfile_T *, blocknr_T));
+static void mf_ins_used __ARGS((memfile_T *, bhdr_T *));
+static void mf_rem_used __ARGS((memfile_T *, bhdr_T *));
+static bhdr_T *mf_release __ARGS((memfile_T *, int));
+static bhdr_T *mf_alloc_bhdr __ARGS((memfile_T *, int));
+static void mf_free_bhdr __ARGS((bhdr_T *));
+static void mf_ins_free __ARGS((memfile_T *, bhdr_T *));
+static bhdr_T *mf_rem_free __ARGS((memfile_T *));
+static int  mf_read __ARGS((memfile_T *, bhdr_T *));
+static int  mf_write __ARGS((memfile_T *, bhdr_T *));
+static int  mf_trans_add __ARGS((memfile_T *, bhdr_T *));
+static void mf_do_open __ARGS((memfile_T *, char_u *, int));
 
 /*
  * The functions for using a memfile:
@@ -109,31 +109,31 @@ static void mf_do_open __ARGS((MEMFILE *, char_u *, int));
  */
 
 /*
- * mf_open: open an existing or new memory block file
+ * Open an existing or new memory block file.
  *
  *  fname:	name of file to use (NULL means no file at all)
  *		Note: fname must have been allocated, it is not copied!
  *			If opening the file fails, fname is freed.
- *  trunc_file:	    if TRUE: file should be truncated when opening
+ *  flags:	flags for open() call
  *
  *  If fname != NULL and file cannot be opened, fail.
  *
  * return value: identifier for this memory block file.
  */
-    MEMFILE *
-mf_open(fname, trunc_file)
-    char_u  *fname;
-    int	    trunc_file;
+    memfile_T *
+mf_open(fname, flags)
+    char_u	*fname;
+    int		flags;
 {
-    MEMFILE	    *mfp;
-    int		    i;
-    off_t	    size;
+    memfile_T		*mfp;
+    int			i;
+    off_t		size;
 #if defined(STATFS) && defined(UNIX) && !defined(__QNX__)
 # define USE_FSTATFS
-    struct STATFS   stf;
+    struct STATFS	stf;
 #endif
 
-    if ((mfp = (MEMFILE *)alloc((unsigned)sizeof(MEMFILE))) == NULL)
+    if ((mfp = (memfile_T *)alloc((unsigned)sizeof(memfile_T))) == NULL)
 	return NULL;
 
     if (fname == NULL)	    /* no file for this memfile, use memory only */
@@ -144,7 +144,7 @@ mf_open(fname, trunc_file)
     }
     else
     {
-	mf_do_open(mfp, fname, trunc_file);	/* try to open the file */
+	mf_do_open(mfp, fname, flags);	/* try to open the file */
 
 	/* if the file cannot be opened, return here */
 	if (mfp->mf_fd < 0)
@@ -170,45 +170,65 @@ mf_open(fname, trunc_file)
     /*
      * Try to set the page size equal to the block size of the device.
      * Speeds up I/O a lot.
-     * NOTE: minimal block size depends on size of block 0 data! It's not done
-     * with a sizeof(), because block 0 is defined in memline.c (Sorry).
-     * The maximal block size is arbitrary.
+     * When recovering, the actual block size will be retrieved from block 0
+     * in ml_recover().  The size used here may be wrong, therefore
+     * mf_blocknr_max must be rounded up.
      */
-    if (mfp->mf_fd >= 0 &&
-		    fstatfs(mfp->mf_fd, &stf, sizeof(struct statfs), 0) == 0 &&
-		    stf.F_BSIZE >= 1048 && stf.F_BSIZE <= 50000)
+    if (mfp->mf_fd >= 0
+	    && fstatfs(mfp->mf_fd, &stf, sizeof(struct statfs), 0) == 0
+	    && stf.F_BSIZE >= MIN_SWAP_PAGE_SIZE
+	    && stf.F_BSIZE <= MAX_SWAP_PAGE_SIZE)
 	mfp->mf_page_size = stf.F_BSIZE;
 #endif
 
-    if (mfp->mf_fd < 0 || trunc_file ||
-			 (size = lseek(mfp->mf_fd, (off_t)0L, SEEK_END)) <= 0)
+    if (mfp->mf_fd < 0 || (flags & (O_TRUNC|O_EXCL))
+		      || (size = lseek(mfp->mf_fd, (off_t)0L, SEEK_END)) <= 0)
 	mfp->mf_blocknr_max = 0;	/* no file or empty file */
     else
-	mfp->mf_blocknr_max = size / mfp->mf_page_size;
+	mfp->mf_blocknr_max = (blocknr_T)((size + mfp->mf_page_size - 1)
+							 / mfp->mf_page_size);
     mfp->mf_blocknr_min = -1;
     mfp->mf_neg_count = 0;
     mfp->mf_infile_count = mfp->mf_blocknr_max;
-    mfp->mf_used_count_max = p_mm * 1024 / mfp->mf_page_size;
+
+    /*
+     * Compute maximum number of pages ('maxmem' is in Kbyte):
+     *	'mammem' * 1Kbyte / page-size-in-bytes.
+     * Avoid overflow by first reducing page size as much as possible.
+     */
+    {
+	int	    shift = 10;
+	unsigned    page_size = mfp->mf_page_size;
+
+	while (shift > 0 && (page_size & 1) == 0)
+	{
+	    page_size = page_size >> 1;
+	    --shift;
+	}
+	mfp->mf_used_count_max = (p_mm << shift) / page_size;
+	if (mfp->mf_used_count_max < 10)
+	    mfp->mf_used_count_max = 10;
+    }
 
     return mfp;
 }
 
 /*
- * mf_open_file: open a file for an existing memfile. Used when updatecount
- *		 set from 0 to some value.
- *
- *  fname:	name of file to use (NULL means no file at all)
- *		Note: fname must have been allocated, it is not copied!
- *			If opening the file fails, fname is freed.
+ * Open a file for an existing memfile.  Used when updatecount set from 0 to
+ * some value.
+ * If the file already exists, this fails.
+ * "fname" is the name of file to use (NULL means no file at all)
+ * Note: "fname" must have been allocated, it is not copied!  If opening the
+ * file fails, "fname" is freed.
  *
  * return value: FAIL if file could not be opened, OK otherwise
  */
     int
 mf_open_file(mfp, fname)
-    MEMFILE	*mfp;
+    memfile_T	*mfp;
     char_u	*fname;
 {
-    mf_do_open(mfp, fname, TRUE);		/* try to open the file */
+    mf_do_open(mfp, fname, O_RDWR|O_CREAT|O_EXCL); /* try to open the file */
 
     if (mfp->mf_fd < 0)
 	return FAIL;
@@ -222,10 +242,10 @@ mf_open_file(mfp, fname)
  */
     void
 mf_close(mfp, del_file)
-    MEMFILE *mfp;
-    int	    del_file;
+    memfile_T	*mfp;
+    int		del_file;
 {
-    BHDR	*hp, *nextp;
+    bhdr_T	*hp, *nextp;
     NR_TRANS	*tp, *tpnext;
     int		i;
 
@@ -234,7 +254,7 @@ mf_close(mfp, del_file)
     if (mfp->mf_fd >= 0)
     {
 	if (close(mfp->mf_fd) < 0)
-	    EMSG("Close error on swap file");
+	    EMSG(_(e_swapclose));
     }
     if (del_file && mfp->mf_fname != NULL)
 	mch_remove(mfp->mf_fname);
@@ -262,26 +282,29 @@ mf_close(mfp, del_file)
  * Close the swap file for a memfile.  Used when 'swapfile' is reset.
  */
     void
-mf_close_file(buf)
-    BUF		*buf;
+mf_close_file(buf, getlines)
+    buf_T	*buf;
+    int		getlines;	/* get all lines into memory? */
 {
-    MEMFILE	*mfp;
-    linenr_t	lnum;
+    memfile_T	*mfp;
+    linenr_T	lnum;
 
     mfp = buf->b_ml.ml_mfp;
     if (mfp == NULL || mfp->mf_fd < 0)		/* nothing to close */
 	return;
 
-    /* get all blocks in memory by accessing all lines (clumsy!) */
-    dont_release = TRUE;
-    for (lnum = 1; lnum <= buf->b_ml.ml_line_count; ++lnum)
-	(void)ml_get_buf(buf, lnum, FALSE);
-    dont_release = FALSE;
-
-    /* TODO: should check if all blocks are really in core */
+    if (getlines)
+    {
+	/* get all blocks in memory by accessing all lines (clumsy!) */
+	mf_dont_release = TRUE;
+	for (lnum = 1; lnum <= buf->b_ml.ml_line_count; ++lnum)
+	    (void)ml_get_buf(buf, lnum, FALSE);
+	mf_dont_release = FALSE;
+	/* TODO: should check if all blocks are really in core */
+    }
 
     if (close(mfp->mf_fd) < 0)			/* close the file */
-	EMSG("Close error on swap file");
+	EMSG(_(e_swapclose));
     mfp->mf_fd = -1;
 
     if (mfp->mf_fname != NULL)
@@ -294,24 +317,39 @@ mf_close_file(buf)
     }
 }
 
+/*
+ * Set new size for a memfile.  Used when block 0 of a swapfile has been read
+ * and the size it indicates differs from what was guessed.
+ */
+    void
+mf_new_page_size(mfp, new_size)
+    memfile_T	*mfp;
+    unsigned	new_size;
+{
+    /* Correct the memory used for block 0 to the new size, because it will be
+     * freed with that size later on. */
+    total_mem_used += new_size - mfp->mf_page_size;
+    mfp->mf_page_size = new_size;
+}
+
 /*
  * get a new block
  *
  *   negative: TRUE if negative block number desired (data block)
  */
-    BHDR *
+    bhdr_T *
 mf_new(mfp, negative, page_count)
-    MEMFILE	*mfp;
+    memfile_T	*mfp;
     int		negative;
     int		page_count;
 {
-    BHDR    *hp;	    /* new BHDR */
-    BHDR    *freep;	    /* first block in free list */
-    char_u  *p;
+    bhdr_T	*hp;	/* new bhdr_T */
+    bhdr_T	*freep;	/* first block in free list */
+    char_u	*p;
 
     /*
      * If we reached the maximum size for the used memory blocks, release one
-     * If a BHDR is returned, use it and adjust the page_count if necessary.
+     * If a bhdr_T is returned, use it and adjust the page_count if necessary.
      */
     hp = mf_release(mfp, page_count);
 
@@ -326,13 +364,13 @@ mf_new(mfp, negative, page_count)
     {
 	/*
 	 * If the block in the free list has more pages, take only the number
-	 * of pages needed and allocate a new BHDR with data
+	 * of pages needed and allocate a new bhdr_T with data
 	 *
-	 * If the number of pages matches and mf_release did not return a BHDR,
-	 * use the BHDR from the free list and allocate the data
+	 * If the number of pages matches and mf_release() did not return a
+	 * bhdr_T, use the bhdr_T from the free list and allocate the data
 	 *
-	 * If the number of pages matches and mf_release returned a BHDR,
-	 * just use the number and free the BHDR from the free list
+	 * If the number of pages matches and mf_release() returned a bhdr_T,
+	 * just use the number and free the bhdr_T from the free list
 	 */
 	if (freep->bh_page_count > page_count)
 	{
@@ -391,13 +429,13 @@ mf_new(mfp, negative, page_count)
  *
  * Note: The caller should first check a negative nr with mf_trans_del()
  */
-    BHDR *
+    bhdr_T *
 mf_get(mfp, nr, page_count)
-    MEMFILE	*mfp;
-    blocknr_t	nr;
+    memfile_T	*mfp;
+    blocknr_T	nr;
     int		page_count;
 {
-    BHDR    *hp;
+    bhdr_T    *hp;
 						/* doesn't exist */
     if (nr >= mfp->mf_blocknr_max || nr <= mfp->mf_blocknr_min)
 	return NULL;
@@ -454,17 +492,17 @@ mf_get(mfp, nr, page_count)
  */
     void
 mf_put(mfp, hp, dirty, infile)
-    MEMFILE *mfp;
-    BHDR    *hp;
-    int	    dirty;
-    int	    infile;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
+    int		dirty;
+    int		infile;
 {
-    int	    flags;
+    int		flags;
 
     flags = hp->bh_flags;
 
     if ((flags & BH_LOCKED) == 0)
-	EMSG("block was not locked");
+	EMSG(_("E293: block was not locked"));
     flags &= ~BH_LOCKED;
     if (dirty)
     {
@@ -481,8 +519,8 @@ mf_put(mfp, hp, dirty, infile)
  */
     void
 mf_free(mfp, hp)
-    MEMFILE *mfp;
-    BHDR    *hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
     vim_free(hp->bh_data);	/* free the memory */
     mf_rem_hash(mfp, hp);	/* get *hp out of the hash list */
@@ -496,6 +534,17 @@ mf_free(mfp, hp)
 	mf_ins_free(mfp, hp);	/* put *hp in the free list */
 }
 
+#if defined(__MORPHOS__) && defined(__libnix__)
+/* function is missing in MorphOS libnix version */
+extern unsigned long *__stdfiledes;
+
+    static unsigned long
+fdtofh(int filedescriptor)
+{
+    return __stdfiledes[filedescriptor];
+}
+#endif
+
 /*
  * Sync the memory file *mfp to disk.
  * Flags:
@@ -511,14 +560,15 @@ mf_free(mfp, hp)
  */
     int
 mf_sync(mfp, flags)
-    MEMFILE *mfp;
-    int	    flags;
+    memfile_T	*mfp;
+    int		flags;
 {
-    int	    status;
-    BHDR    *hp;
-#ifdef SYNC_DUP_CLOSE
-    int	    fd;
+    int		status;
+    bhdr_T	*hp;
+#if defined(SYNC_DUP_CLOSE) && !defined(MSDOS)
+    int		fd;
 #endif
+    int		got_int_save = got_int;
 
     if (mfp->mf_fd < 0)	    /* there is no file, nothing to do */
     {
@@ -526,6 +576,10 @@ mf_sync(mfp, flags)
 	return FAIL;
     }
 
+    /* Only a CTRL-C while writing will break us here, not one typed
+     * previously. */
+    got_int = FALSE;
+
     /*
      * sync from last to first (may reduce the probability of an inconsistent
      * file) If a write fails, it is very likely caused by a full filesystem.
@@ -547,8 +601,15 @@ mf_sync(mfp, flags)
 		    break;
 		status = FAIL;
 	    }
-	    /* Stop when char available now */
-	    if ((flags & MFS_STOP) && ui_char_avail())
+	    if (flags & MFS_STOP)
+	    {
+		/* Stop when char available now. */
+		if (ui_char_avail())
+		    break;
+	    }
+	    else
+		ui_breakcheck();
+	    if (got_int)
 		break;
 	}
 
@@ -582,8 +643,10 @@ mf_sync(mfp, flags)
 	}
 	else
 # endif
-# ifdef __OPENNT
-	    fflush(NULL);	/* OpenNT is strictly POSIX (Benzinger) */
+	    /* OpenNT is strictly POSIX (Benzinger) */
+	    /* Tandem/Himalaya NSK-OSS doesn't have sync() */
+# if defined(__OPENNT) || defined(__TANDEM)
+	    fflush(NULL);
 # else
 	    sync();
 # endif
@@ -609,49 +672,82 @@ mf_sync(mfp, flags)
 # endif
 #endif
 #ifdef AMIGA
+# if defined(__AROS__) || defined(__amigaos4__)
+	if (fsync(mfp->mf_fd) != 0)
+	    status = FAIL;
+# else
 	/*
 	 * Flush() only exists for AmigaDos 2.0.
 	 * For 1.3 it should be done with close() + open(), but then the risk
 	 * is that the open() may fail and lose the file....
 	 */
-# ifndef NO_ARP
+#  ifdef FEAT_ARP
 	if (dos2)
-# endif
-# ifdef SASC
+#  endif
+#  ifdef SASC
 	{
 	    struct UFB *fp = chkufb(mfp->mf_fd);
 
 	    if (fp != NULL)
 		Flush(fp->ufbfh);
 	}
-# else
-#  ifdef _DCC
+#  else
+#   if defined(_DCC) || defined(__GNUC__) || defined(__MORPHOS__)
 	{
+#    if defined(__GNUC__) && !defined(__MORPHOS__) && defined(__libnix__)
+	    /* Have function (in libnix at least),
+	     * but ain't got no prototype anywhere. */
+	    extern unsigned long fdtofh(int filedescriptor);
+#    endif
+#    if !defined(__libnix__)
+	    fflush(NULL);
+#    else
 	    BPTR fh = (BPTR)fdtofh(mfp->mf_fd);
 
 	    if (fh != 0)
 		Flush(fh);
-	    }
-#  else /* assume Manx */
+#    endif
+	}
+#   else /* assume Manx */
 	    Flush(_devtab[mfp->mf_fd].fd);
+#   endif
 #  endif
 # endif
 #endif /* AMIGA */
     }
 
+    got_int |= got_int_save;
+
     return status;
 }
 
+/*
+ * For all blocks in memory file *mfp that have a positive block number set
+ * the dirty flag.  These are blocks that need to be written to a newly
+ * created swapfile.
+ */
+    void
+mf_set_dirty(mfp)
+    memfile_T	*mfp;
+{
+    bhdr_T	*hp;
+
+    for (hp = mfp->mf_used_last; hp != NULL; hp = hp->bh_prev)
+	if (hp->bh_bnum > 0)
+	    hp->bh_flags |= BH_DIRTY;
+    mfp->mf_dirty = TRUE;
+}
+
 /*
  * insert block *hp in front of hashlist of memfile *mfp
  */
     static void
 mf_ins_hash(mfp, hp)
-    MEMFILE *mfp;
-    BHDR    *hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
-    BHDR    *hhp;
-    int	    hash;
+    bhdr_T	*hhp;
+    int		hash;
 
     hash = MEMHASH(hp->bh_bnum);
     hhp = mfp->mf_hash[hash];
@@ -667,8 +763,8 @@ mf_ins_hash(mfp, hp)
  */
     static void
 mf_rem_hash(mfp, hp)
-    MEMFILE *mfp;
-    BHDR    *hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
     if (hp->bh_hash_prev == NULL)
 	mfp->mf_hash[MEMHASH(hp->bh_bnum)] = hp->bh_hash_next;
@@ -682,12 +778,12 @@ mf_rem_hash(mfp, hp)
 /*
  * look in hash lists of memfile *mfp for block header with number 'nr'
  */
-    static BHDR *
+    static bhdr_T *
 mf_find_hash(mfp, nr)
-    MEMFILE	*mfp;
-    blocknr_t	nr;
+    memfile_T	*mfp;
+    blocknr_T	nr;
 {
-    BHDR	*hp;
+    bhdr_T	*hp;
 
     for (hp = mfp->mf_hash[MEMHASH(nr)]; hp != NULL; hp = hp->bh_hash_next)
 	if (hp->bh_bnum == nr)
@@ -700,8 +796,8 @@ mf_find_hash(mfp, nr)
  */
     static void
 mf_ins_used(mfp, hp)
-    MEMFILE *mfp;
-    BHDR    *hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
     hp->bh_next = mfp->mf_used_first;
     mfp->mf_used_first = hp;
@@ -719,8 +815,8 @@ mf_ins_used(mfp, hp)
  */
     static void
 mf_rem_used(mfp, hp)
-    MEMFILE *mfp;
-    BHDR    *hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
     if (hp->bh_next == NULL)	    /* last block in used list */
 	mfp->mf_used_last = hp->bh_prev;
@@ -741,17 +837,17 @@ mf_rem_used(mfp, hp)
  * Return the block header to the caller, including the memory block, so
  * it can be re-used. Make sure the page_count is right.
  */
-    static BHDR *
+    static bhdr_T *
 mf_release(mfp, page_count)
-    MEMFILE	*mfp;
+    memfile_T	*mfp;
     int		page_count;
 {
-    BHDR	*hp;
+    bhdr_T	*hp;
     int		need_release;
-    BUF		*buf;
+    buf_T	*buf;
 
     /* don't release while in mf_close_file() */
-    if (dont_release)
+    if (mf_dont_release)
 	return NULL;
 
     /*
@@ -803,7 +899,8 @@ mf_release(mfp, page_count)
     mf_rem_hash(mfp, hp);
 
     /*
-     * If a BHDR is returned, make sure that the page_count of bh_data is right
+     * If a bhdr_T is returned, make sure that the page_count of bh_data is
+     * right
      */
     if (hp->bh_page_count != page_count)
     {
@@ -827,30 +924,40 @@ mf_release(mfp, page_count)
     int
 mf_release_all()
 {
-    BUF		*buf;
-    MEMFILE	*mfp;
-    BHDR	*hp;
+    buf_T	*buf;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
     int		retval = FALSE;
 
     for (buf = firstbuf; buf != NULL; buf = buf->b_next)
     {
 	mfp = buf->b_ml.ml_mfp;
-	/* only if there is a memfile with a file */
-	if (mfp != NULL && mfp->mf_fd >= 0)
-	    for (hp = mfp->mf_used_last; hp != NULL; )
+	if (mfp != NULL)
+	{
+	    /* If no swap file yet, may open one */
+	    if (mfp->mf_fd < 0 && buf->b_may_swap)
+		ml_open_file(buf);
+
+	    /* only if there is a swapfile */
+	    if (mfp->mf_fd >= 0)
 	    {
-		if (!(hp->bh_flags & BH_LOCKED) && (!(hp->bh_flags & BH_DIRTY)
-						|| mf_write(mfp, hp) != FAIL))
+		for (hp = mfp->mf_used_last; hp != NULL; )
 		{
-		    mf_rem_used(mfp, hp);
-		    mf_rem_hash(mfp, hp);
-		    mf_free_bhdr(hp);
-		    hp = mfp->mf_used_last;	/* re-start, list was changed */
-		    retval = TRUE;
+		    if (!(hp->bh_flags & BH_LOCKED)
+			    && (!(hp->bh_flags & BH_DIRTY)
+				|| mf_write(mfp, hp) != FAIL))
+		    {
+			mf_rem_used(mfp, hp);
+			mf_rem_hash(mfp, hp);
+			mf_free_bhdr(hp);
+			hp = mfp->mf_used_last;	/* re-start, list was changed */
+			retval = TRUE;
+		    }
+		    else
+			hp = hp->bh_prev;
 		}
-		else
-		    hp = hp->bh_prev;
 	    }
+	}
     }
     return retval;
 }
@@ -858,14 +965,14 @@ mf_release_all()
 /*
  * Allocate a block header and a block of memory for it
  */
-    static BHDR *
+    static bhdr_T *
 mf_alloc_bhdr(mfp, page_count)
-    MEMFILE	*mfp;
+    memfile_T	*mfp;
     int		page_count;
 {
-    BHDR    *hp;
+    bhdr_T	*hp;
 
-    if ((hp = (BHDR *)alloc((unsigned)sizeof(BHDR))) != NULL)
+    if ((hp = (bhdr_T *)alloc((unsigned)sizeof(bhdr_T))) != NULL)
     {
 	if ((hp->bh_data = (char_u *)alloc(mfp->mf_page_size * page_count))
 								      == NULL)
@@ -883,7 +990,7 @@ mf_alloc_bhdr(mfp, page_count)
  */
     static void
 mf_free_bhdr(hp)
-    BHDR	*hp;
+    bhdr_T	*hp;
 {
     vim_free(hp->bh_data);
     vim_free(hp);
@@ -894,8 +1001,8 @@ mf_free_bhdr(hp)
  */
     static void
 mf_ins_free(mfp, hp)
-    MEMFILE *mfp;
-    BHDR    *hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
     hp->bh_next = mfp->mf_free_first;
     mfp->mf_free_first = hp;
@@ -905,11 +1012,11 @@ mf_ins_free(mfp, hp)
  * remove the first entry from the free list and return a pointer to it
  * Note: caller must check that mfp->mf_free_first is not NULL!
  */
-    static BHDR *
+    static bhdr_T *
 mf_rem_free(mfp)
-    MEMFILE *mfp;
+    memfile_T	*mfp;
 {
-    BHDR    *hp;
+    bhdr_T	*hp;
 
     hp = mfp->mf_free_first;
     mfp->mf_free_first = hp->bh_next;
@@ -923,8 +1030,8 @@ mf_rem_free(mfp)
  */
     static int
 mf_read(mfp, hp)
-    MEMFILE	*mfp;
-    BHDR	*hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
     off_t	offset;
     unsigned	page_size;
@@ -934,16 +1041,16 @@ mf_read(mfp, hp)
 	return FAIL;
 
     page_size = mfp->mf_page_size;
-    offset = page_size * hp->bh_bnum;
+    offset = (off_t)page_size * hp->bh_bnum;
     size = page_size * hp->bh_page_count;
     if (lseek(mfp->mf_fd, offset, SEEK_SET) != offset)
     {
-	EMSG("Seek error in swap file read");
+	PERROR(_("E294: Seek error in swap file read"));
 	return FAIL;
     }
-    if ((unsigned)read(mfp->mf_fd, (char *)hp->bh_data, (size_t)size) != size)
+    if ((unsigned)vim_read(mfp->mf_fd, hp->bh_data, size) != size)
     {
-	EMSG("Read error in swap file");
+	PERROR(_("E295: Read error in swap file"));
 	return FAIL;
     }
     return OK;
@@ -956,12 +1063,12 @@ mf_read(mfp, hp)
  */
     static int
 mf_write(mfp, hp)
-    MEMFILE	*mfp;
-    BHDR	*hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
     off_t	offset;	    /* offset in the file */
-    blocknr_t	nr;	    /* block nr which is being written */
-    BHDR	*hp2;
+    blocknr_T	nr;	    /* block nr which is being written */
+    bhdr_T	*hp2;
     unsigned	page_size;  /* number of bytes in a page */
     unsigned	page_count; /* number of pages written */
     unsigned	size;	    /* number of bytes written */
@@ -992,33 +1099,28 @@ mf_write(mfp, hp)
 	else
 	    hp2 = hp;
 
-	offset = page_size * nr;
-// Ȃlseek()̖߂loffset̔rG[ɂȂ̂ŉ}[u 060617
-{
-	int rest = lseek(mfp->mf_fd, offset, SEEK_SET);
-	if (rest != offset)
-//	if (lseek(mfp->mf_fd, offset, SEEK_SET) != offset)
+	offset = (off_t)page_size * nr;
+	if (lseek(mfp->mf_fd, offset, SEEK_SET) != offset)
 	{
-	    EMSG("Seek error in swap file write");
+	    PERROR(_("E296: Seek error in swap file write"));
 	    return FAIL;
 	}
-}
 	if (hp2 == NULL)	    /* freed block, fill with dummy data */
 	    page_count = 1;
 	else
 	    page_count = hp2->bh_page_count;
 	size = page_size * page_count;
-	if ((unsigned)write(mfp->mf_fd,
-	     (char *)(hp2 == NULL ? hp : hp2)->bh_data, (size_t)size) != size)
+	if ((unsigned)vim_write(mfp->mf_fd,
+	     (hp2 == NULL ? hp : hp2)->bh_data, size) != size)
 	{
 	    /*
 	     * Avoid repeating the error message, this mostly happens when the
-	     * disk is full. We give the message again only after a succesful
+	     * disk is full. We give the message again only after a successful
 	     * write or when hitting a key. We keep on trying, in case some
 	     * space becomes available.
 	     */
 	    if (!did_swapwrite_msg)
-		EMSG("Write error in swap file");
+		EMSG(_("E297: Write error in swap file"));
 	    did_swapwrite_msg = TRUE;
 	    return FAIL;
 	}
@@ -1026,7 +1128,7 @@ mf_write(mfp, hp)
 	if (hp2 != NULL)		    /* written a non-dummy block */
 	    hp2->bh_flags &= ~BH_DIRTY;
 					    /* appended to the file */
-	if (nr + (blocknr_t)page_count > mfp->mf_infile_count)
+	if (nr + (blocknr_T)page_count > mfp->mf_infile_count)
 	    mfp->mf_infile_count = nr + page_count;
 	if (nr == hp->bh_bnum)		    /* written the desired block */
 	    break;
@@ -1041,11 +1143,11 @@ mf_write(mfp, hp)
  */
     static int
 mf_trans_add(mfp, hp)
-    MEMFILE *mfp;
-    BHDR    *hp;
+    memfile_T	*mfp;
+    bhdr_T	*hp;
 {
-    BHDR	*freep;
-    blocknr_t	new_bnum;
+    bhdr_T	*freep;
+    blocknr_T	new_bnum;
     int		hash;
     NR_TRANS	*np;
     int		page_count;
@@ -1105,18 +1207,18 @@ mf_trans_add(mfp, hp)
 }
 
 /*
- * Lookup a tranlation from the trans lists and delete the entry
+ * Lookup a translation from the trans lists and delete the entry
  *
  * Return the positive new number when found, the old number when not found
  */
-    blocknr_t
+    blocknr_T
 mf_trans_del(mfp, old_nr)
-    MEMFILE	*mfp;
-    blocknr_t	old_nr;
+    memfile_T	*mfp;
+    blocknr_T	old_nr;
 {
     int		hash;
     NR_TRANS	*np;
-    blocknr_t	new_bnum;
+    blocknr_T	new_bnum;
 
     hash = MEMHASH(old_nr);
     for (np = mfp->mf_trans[hash]; np != NULL; np = np->nt_next)
@@ -1145,7 +1247,7 @@ mf_trans_del(mfp, old_nr)
  */
     void
 mf_set_ffname(mfp)
-    MEMFILE	*mfp;
+    memfile_T	*mfp;
 {
     mfp->mf_ffname = FullName_save(mfp->mf_fname, FALSE);
 }
@@ -1156,7 +1258,7 @@ mf_set_ffname(mfp)
  */
     void
 mf_fullname(mfp)
-    MEMFILE	*mfp;
+    memfile_T	*mfp;
 {
     if (mfp != NULL && mfp->mf_fname != NULL && mfp->mf_ffname != NULL)
     {
@@ -1171,7 +1273,7 @@ mf_fullname(mfp)
  */
     int
 mf_need_trans(mfp)
-    MEMFILE	*mfp;
+    memfile_T	*mfp;
 {
     return (mfp->mf_fname != NULL && mfp->mf_neg_count > 0);
 }
@@ -1182,19 +1284,24 @@ mf_need_trans(mfp)
  * error occurs).
  */
     static void
-mf_do_open(mfp, fname, trunc_file)
-    MEMFILE	*mfp;
+mf_do_open(mfp, fname, flags)
+    memfile_T	*mfp;
     char_u	*fname;
-    int		trunc_file;
+    int		flags;		/* flags for open() */
 {
+#ifdef HAVE_LSTAT
+    struct stat sb;
+#endif
+
     mfp->mf_fname = fname;
+
     /*
      * Get the full path name before the open, because this is
      * not possible after the open on the Amiga.
      * fname cannot be NameBuff, because it must have been allocated.
      */
     mf_set_ffname(mfp);
-#if defined(MSDOS) || defined(WIN32) || defined(RISCOS)
+#if defined(MSDOS) || defined(MSWIN) || defined(RISCOS)
     /*
      * A ":!cd e:xxx" may change the directory without us knowning, use the
      * full pathname always.  Careful: This frees fname!
@@ -1202,21 +1309,31 @@ mf_do_open(mfp, fname, trunc_file)
     mf_fullname(mfp);
 #endif
 
+#ifdef HAVE_LSTAT
     /*
-     * try to open the file
+     * Extra security check: When creating a swap file it really shouldn't
+     * exist yet.  If there is a symbolic link, this is most likely an attack.
      */
-    mfp->mf_fd = open((char *)mfp->mf_fname,
-	    (trunc_file ? (O_CREAT | O_RDWR | O_TRUNC) : (O_RDONLY)) | O_EXTRA
-#if defined(UNIX) || defined(RISCOS)		 /* open in rw------- mode */
-		    , (mode_t)0600
+    if ((flags & O_CREAT) && mch_lstat((char *)mfp->mf_fname, &sb) >= 0)
+    {
+	mfp->mf_fd = -1;
+	EMSG(_("E300: Swap file already exists (symlink attack?)"));
+    }
+    else
 #endif
-#if defined(MSDOS) || defined(WIN32) || defined(__EMX__)
-		    , S_IREAD | S_IWRITE	 /* open read/write */
+    {
+	/*
+	 * try to open the file
+	 */
+	flags |= O_EXTRA | O_NOFOLLOW;
+#ifdef WIN32
+	/* Prevent handle inheritance that cause problems with Cscope
+	 * (swap file may not be deleted if cscope connection was open after
+	 * the file) */
+	flags |= O_NOINHERIT;
 #endif
-#ifdef VMS		    /* open in rw------- mode */
-		    , 0600
-#endif
-		    );
+	mfp->mf_fd = mch_open_rw((char *)mfp->mf_fname, flags);
+    }
 
     /*
      * If the file cannot be opened, use memory only
@@ -1229,5 +1346,10 @@ mf_do_open(mfp, fname, trunc_file)
 	mfp->mf_ffname = NULL;
     }
     else
+    {
+#ifdef HAVE_SELINUX
+	mch_copy_sec(fname, mfp->mf_fname);
+#endif
 	mch_hide(mfp->mf_fname);    /* try setting the 'hidden' flag */
+    }
 }
