@@ -9,10 +9,15 @@
 
 #include "vim.h"
 
+#ifdef AMIGA
+# include <time.h>	/* for time() */
+#endif
+
 /*
  * Vim originated from Stevie version 3.6 (Fish disk 217) by GRWalter (Fred)
  * It has been changed beyond recognition since then.
  *
+ * Differences between version 6.x and 7.x can be found with ":help version7".
  * Differences between version 5.x and 6.x can be found with ":help version6".
  * Differences between version 4.x and 5.x can be found with ":help version5".
  * Differences between version 3.0 and 4.x can be found with ":help version4".
@@ -22,8 +27,8 @@
 
 #include "version.h"
 
-char	*Version = VIM_VERSION_SHORT;
-char	*mediumVersion = VIM_VERSION_MEDIUM;
+char		*Version = VIM_VERSION_SHORT;
+static char	*mediumVersion = VIM_VERSION_MEDIUM;
 
 #if defined(HAVE_DATE_TIME) || defined(PROTO)
 # if (defined(VMS) && defined(VAXC)) || defined(PROTO)
@@ -60,6 +65,11 @@ static char *(features[]) =
 	"-ARP",
 # endif
 #endif
+#ifdef FEAT_ARABIC
+	"+arabic",
+#else
+	"-arabic",
+#endif
 #ifdef FEAT_AUTOCMD
 	"+autocmd",
 #else
@@ -134,6 +144,11 @@ static char *(features[]) =
 #else
 	"-cscope",
 #endif
+#ifdef CURSOR_SHAPE
+	"+cursorshape",
+#else
+	"-cursorshape",
+#endif
 #if defined(FEAT_CON_DIALOG) && defined(FEAT_GUI_DIALOG)
 	"+dialog_con_gui",
 #else
@@ -157,6 +172,11 @@ static char *(features[]) =
 #else
 	"-digraphs",
 #endif
+#ifdef FEAT_DND
+	"+dnd",
+#else
+	"-dnd",
+#endif
 #ifdef EBCDIC
 	"+ebcdic",
 #else
@@ -356,6 +376,20 @@ static char *(features[]) =
 #else
 	"-multi_lang",
 #endif
+#ifdef FEAT_MZSCHEME
+# ifdef DYNAMIC_MZSCHEME
+	"+mzscheme/dyn",
+# else
+	"+mzscheme",
+# endif
+#else
+	"-mzscheme",
+#endif
+#ifdef FEAT_NETBEANS_INTG
+	"+netbeans_intg",
+#else
+	"-netbeans_intg",
+#endif
 #ifdef FEAT_GUI_W32
 # ifdef FEAT_OLE
 	"+ole",
@@ -392,6 +426,11 @@ static char *(features[]) =
 #else
 	"-printer",
 #endif
+#ifdef FEAT_PROFILE
+	"+profile",
+#else
+	"-profile",
+#endif
 #ifdef FEAT_PYTHON
 # ifdef DYNAMIC_PYTHON
 	"+python/dyn",
@@ -406,6 +445,11 @@ static char *(features[]) =
 #else
 	"-quickfix",
 #endif
+#ifdef FEAT_RELTIME
+	"+reltime",
+#else
+	"-reltime",
+#endif
 #ifdef FEAT_RIGHTLEFT
 	"+rightleft",
 #else
@@ -590,6 +634,15 @@ static char *(features[]) =
 	"-xim",
 #endif
 #if defined(UNIX) || defined(VMS)
+# ifdef USE_XSMP_INTERACT
+	"+xsmp_interact",
+# else
+#  ifdef USE_XSMP
+	"+xsmp",
+#  else
+	"-xsmp",
+#  endif
+# endif
 # ifdef FEAT_XCLIPBOARD
 	"+xterm_clipboard",
 # else
@@ -600,12 +653,611 @@ static char *(features[]) =
 	"+xterm_save",
 #else
 	"-xterm_save",
+#endif
+#ifdef WIN3264
+# ifdef FEAT_XPM_W32
+	"+xpm_w32",
+# else
+	"-xpm_w32",
+# endif
 #endif
 	NULL
 };
 
 static int included_patches[] =
 {   /* Add new patch number below this line */
+/**/
+    314,
+/**/
+    313,
+/**/
+    312,
+/**/
+    311,
+/**/
+    310,
+/**/
+    309,
+/**/
+    308,
+/**/
+    307,
+/**/
+    306,
+/**/
+    305,
+/**/
+    304,
+/**/
+    303,
+/**/
+    302,
+/**/
+    301,
+/**/
+    300,
+/**/
+    299,
+/**/
+    298,
+/**/
+    297,
+/**/
+    296,
+/**/
+    295,
+/**/
+    294,
+/**/
+    293,
+/**/
+    292,
+/**/
+    291,
+/**/
+    290,
+/**/
+    289,
+/**/
+    288,
+/**/
+    287,
+/**/
+    286,
+/**/
+    285,
+/**/
+    284,
+/**/
+    283,
+/**/
+    282,
+/**/
+    281,
+/**/
+    280,
+/**/
+    279,
+/**/
+    278,
+/**/
+    277,
+/**/
+    276,
+/**/
+    275,
+/**/
+    274,
+/**/
+    273,
+/**/
+    272,
+/**/
+    271,
+/**/
+    270,
+/**/
+    269,
+/**/
+    268,
+/**/
+    267,
+/**/
+    266,
+/**/
+    265,
+/**/
+    264,
+/**/
+    263,
+/**/
+    262,
+/**/
+    261,
+/**/
+    260,
+/**/
+    259,
+/**/
+    258,
+/**/
+    257,
+/**/
+    256,
+/**/
+    255,
+/**/
+    254,
+/**/
+    253,
+/**/
+    252,
+/**/
+    251,
+/**/
+    250,
+/**/
+    249,
+/**/
+    248,
+/**/
+    247,
+/**/
+    246,
+/**/
+    245,
+/**/
+    244,
+/**/
+    243,
+/**/
+    242,
+/**/
+    241,
+/**/
+    240,
+/**/
+    239,
+/**/
+    238,
+/**/
+    237,
+/**/
+    236,
+/**/
+    235,
+/**/
+    234,
+/**/
+    233,
+/**/
+    232,
+/**/
+    231,
+/**/
+    230,
+/**/
+    229,
+/**/
+    228,
+/**/
+    227,
+/**/
+    226,
+/**/
+    225,
+/**/
+    224,
+/**/
+    223,
+/**/
+    222,
+/**/
+    221,
+/**/
+    220,
+/**/
+    219,
+/**/
+    218,
+/**/
+    217,
+/**/
+    216,
+/**/
+    215,
+/**/
+    214,
+/**/
+    213,
+/**/
+    212,
+/**/
+    211,
+/**/
+    210,
+/**/
+    209,
+/**/
+    208,
+/**/
+    207,
+/**/
+    206,
+/**/
+    205,
+/**/
+    204,
+/**/
+    203,
+/**/
+    202,
+/**/
+    201,
+/**/
+    200,
+/**/
+    199,
+/**/
+    198,
+/**/
+    197,
+/**/
+    196,
+/**/
+    195,
+/**/
+    194,
+/**/
+    193,
+/**/
+    192,
+/**/
+    191,
+/**/
+    190,
+/**/
+    189,
+/**/
+    188,
+/**/
+    187,
+/**/
+    186,
+/**/
+    185,
+/**/
+    184,
+/**/
+    183,
+/**/
+    182,
+/**/
+    181,
+/**/
+    180,
+/**/
+    179,
+/**/
+    178,
+/**/
+    177,
+/**/
+    176,
+/**/
+    175,
+/**/
+    174,
+/**/
+    173,
+/**/
+    172,
+/**/
+    171,
+/**/
+    170,
+/**/
+    169,
+/**/
+    168,
+/**/
+    167,
+/**/
+    166,
+/**/
+    165,
+/**/
+    164,
+/**/
+    163,
+/**/
+    162,
+/**/
+    161,
+/**/
+    160,
+/**/
+    159,
+/**/
+    158,
+/**/
+    157,
+/**/
+    156,
+/**/
+    155,
+/**/
+    154,
+/**/
+    153,
+/**/
+    152,
+/**/
+    151,
+/**/
+    150,
+/**/
+    149,
+/**/
+    148,
+/**/
+    147,
+/**/
+    146,
+/**/
+    145,
+/**/
+    144,
+/**/
+    143,
+/**/
+    142,
+/**/
+    141,
+/**/
+    140,
+/**/
+    139,
+/**/
+    138,
+/**/
+    137,
+/**/
+    136,
+/**/
+    135,
+/**/
+    134,
+/**/
+    133,
+/**/
+    132,
+/**/
+    131,
+/**/
+    130,
+/**/
+    129,
+/**/
+    128,
+/**/
+    127,
+/**/
+    126,
+/**/
+    125,
+/**/
+    124,
+/**/
+    123,
+/**/
+    122,
+/**/
+    121,
+/**/
+    120,
+/**/
+    119,
+/**/
+    118,
+/**/
+    117,
+/**/
+    116,
+/**/
+    115,
+/**/
+    114,
+/**/
+    113,
+/**/
+    112,
+/**/
+    111,
+/**/
+    110,
+/**/
+    109,
+/**/
+    108,
+/**/
+    107,
+/**/
+    106,
+/**/
+    105,
+/**/
+    104,
+/**/
+    103,
+/**/
+    102,
+/**/
+    101,
+/**/
+    100,
+/**/
+    99,
+/**/
+    98,
+/**/
+    97,
+/**/
+    96,
+/**/
+    95,
+/**/
+    94,
+/**/
+    93,
+/**/
+    92,
+/**/
+    91,
+/**/
+    90,
+/**/
+    89,
+/**/
+    88,
+/**/
+    87,
+/**/
+    86,
+/**/
+    85,
+/**/
+    84,
+/**/
+    83,
+/**/
+    82,
+/**/
+    81,
+/**/
+    80,
+/**/
+    79,
+/**/
+    78,
+/**/
+    77,
+/**/
+    76,
+/**/
+    75,
+/**/
+    74,
+/**/
+    73,
+/**/
+    72,
+/**/
+    71,
+/**/
+    70,
+/**/
+    69,
+/**/
+    68,
+/**/
+    67,
+/**/
+    66,
+/**/
+    65,
+/**/
+    64,
+/**/
+    63,
+/**/
+    62,
+/**/
+    61,
+/**/
+    60,
+/**/
+    59,
+/**/
+    58,
+/**/
+    57,
+/**/
+    56,
+/**/
+    55,
+/**/
+    54,
+/**/
+    53,
+/**/
+    52,
+/**/
+    51,
+/**/
+    50,
+/**/
+    49,
+/**/
+    48,
+/**/
+    47,
+/**/
+    46,
+/**/
+    45,
+/**/
+    44,
+/**/
+    43,
+/**/
+    42,
+/**/
+    41,
+/**/
+    40,
+/**/
+    39,
+/**/
+    38,
+/**/
+    37,
+/**/
+    36,
+/**/
+    35,
+/**/
+    34,
+/**/
+    33,
+/**/
+    32,
+/**/
+    31,
+/**/
+    30,
+/**/
+    29,
+/**/
+    28,
+/**/
+    27,
+/**/
+    26,
+/**/
+    25,
+/**/
+    24,
+/**/
+    23,
+/**/
+    22,
+/**/
+    21,
+/**/
+    20,
+/**/
+    19,
 /**/
     18,
 /**/
@@ -658,6 +1310,23 @@ highest_patch()
     return h;
 }
 
+#if defined(FEAT_EVAL) || defined(PROTO)
+/*
+ * Return TRUE if patch "n" has been included.
+ */
+    int
+has_patch(n)
+    int		n;
+{
+    int		i;
+
+    for (i = 0; included_patches[i] != 0; ++i)
+	if (included_patches[i] == n)
+	    return TRUE;
+    return FALSE;
+}
+#endif
+
     void
 ex_version(eap)
     exarg_T	*eap;
@@ -690,7 +1359,11 @@ list_version()
     /* Only MS VC 4.1 and earlier can do Win32s */
     MSG_PUTS(_("\nMS-Windows 16/32 bit GUI version"));
 #  else
+#   ifdef _WIN64
+    MSG_PUTS(_("\nMS-Windows 64 bit GUI version"));
+#   else
     MSG_PUTS(_("\nMS-Windows 32 bit GUI version"));
+#   endif
 #  endif
     if (gui_is_win32s())
 	MSG_PUTS(_(" in Win32s mode"));
@@ -728,6 +1401,14 @@ list_version()
 #endif
 #ifdef VMS
     MSG_PUTS("\nOpenVMS version");
+# ifdef HAVE_PATHDEF
+    if (*compiled_arch != NUL)
+    {
+	MSG_PUTS(" - ");
+	MSG_PUTS(compiled_arch);
+    }
+# endif
+
 #endif
 
     /* Print the list of patch numbers if there is at least one. */
@@ -803,20 +1484,29 @@ list_version()
 #else
 # ifdef FEAT_GUI_GTK
 #  ifdef FEAT_GUI_GNOME
+#   ifdef HAVE_GTK2
+    MSG_PUTS(_("with GTK2-GNOME GUI."));
+#   else
     MSG_PUTS(_("with GTK-GNOME GUI."));
+#   endif
 #  else
+#   ifdef HAVE_GTK2
+    MSG_PUTS(_("with GTK2 GUI."));
+#   else
     MSG_PUTS(_("with GTK GUI."));
+#   endif
 #  endif
 # else
 #  ifdef FEAT_GUI_MOTIF
     MSG_PUTS(_("with X11-Motif GUI."));
 #  else
 #   ifdef FEAT_GUI_ATHENA
-    MSG_PUTS(_("with X11-Athena GUI."));
-#   else
-#    ifdef FEAT_GUI_BEOS
-    MSG_PUTS(_("with BeOS GUI."));
+#    ifdef FEAT_GUI_NEXTAW
+    MSG_PUTS(_("with X11-neXtaw GUI."));
 #    else
+    MSG_PUTS(_("with X11-Athena GUI."));
+#    endif
+#   else
 #     ifdef FEAT_GUI_PHOTON
     MSG_PUTS(_("with Photon GUI."));
 #     else
@@ -835,78 +1525,77 @@ list_version()
 #	 endif
 #	endif
 #      endif
-#     endif
 #    endif
 #   endif
 #  endif
 # endif
 #endif
-    MSG_PUTS(_("  Features included (+) or not (-):\n"));
+    version_msg(_("  Features included (+) or not (-):\n"));
 
     /* print all the features */
     for (i = 0; features[i] != NULL; ++i)
     {
 	version_msg(features[i]);
 	if (msg_col > 0)
-	    msg_putchar(' ');
+	    version_msg(" ");
     }
 
-    msg_putchar('\n');
+    version_msg("\n");
 #ifdef SYS_VIMRC_FILE
     version_msg(_("   system vimrc file: \""));
     version_msg(SYS_VIMRC_FILE);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 #endif
 #ifdef USR_VIMRC_FILE
     version_msg(_("     user vimrc file: \""));
     version_msg(USR_VIMRC_FILE);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 #endif
 #ifdef USR_VIMRC_FILE2
     version_msg(_(" 2nd user vimrc file: \""));
     version_msg(USR_VIMRC_FILE2);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 #endif
 #ifdef USR_VIMRC_FILE3
     version_msg(_(" 3rd user vimrc file: \""));
     version_msg(USR_VIMRC_FILE3);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 #endif
 #ifdef USR_EXRC_FILE
     version_msg(_("      user exrc file: \""));
     version_msg(USR_EXRC_FILE);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 #endif
 #ifdef USR_EXRC_FILE2
     version_msg(_("  2nd user exrc file: \""));
     version_msg(USR_EXRC_FILE2);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 #endif
 #ifdef FEAT_GUI
 # ifdef SYS_GVIMRC_FILE
     version_msg(_("  system gvimrc file: \""));
     version_msg(SYS_GVIMRC_FILE);
-    MSG_PUTS(_("\"\n"));
+    version_msg("\"\n");
 # endif
     version_msg(_("    user gvimrc file: \""));
     version_msg(USR_GVIMRC_FILE);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 # ifdef USR_GVIMRC_FILE2
     version_msg(_("2nd user gvimrc file: \""));
     version_msg(USR_GVIMRC_FILE2);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 # endif
 # ifdef USR_GVIMRC_FILE3
     version_msg(_("3rd user gvimrc file: \""));
     version_msg(USR_GVIMRC_FILE3);
-    version_msg(_("\"\n"));
+    version_msg("\"\n");
 # endif
 #endif
 #ifdef FEAT_GUI
 # ifdef SYS_MENU_FILE
     version_msg(_("    system menu file: \""));
     version_msg(SYS_MENU_FILE);
-    MSG_PUTS(_("\"\n"));
+    version_msg("\"\n");
 # endif
 #endif
 #ifdef HAVE_PATHDEF
@@ -914,30 +1603,30 @@ list_version()
     {
 	version_msg(_("  fall-back for $VIM: \""));
 	version_msg((char *)default_vim_dir);
-	MSG_PUTS(_("\"\n"));
+	version_msg("\"\n");
     }
     if (*default_vimruntime_dir != NUL)
     {
 	version_msg(_(" f-b for $VIMRUNTIME: \""));
 	version_msg((char *)default_vimruntime_dir);
-	MSG_PUTS(_("\"\n"));
+	version_msg("\"\n");
     }
     version_msg(_("Compilation: "));
     version_msg((char *)all_cflags);
-    msg_putchar('\n');
+    version_msg("\n");
 #ifdef VMS
     if (*compiler_version != NUL)
     {
 	version_msg(_("Compiler: "));
 	version_msg((char *)compiler_version);
-	msg_putchar('\n');
+	version_msg("\n");
     }
 #endif
     version_msg(_("Linking: "));
     version_msg((char *)all_lflags);
 #endif
 #ifdef DEBUG
-    msg_putchar('\n');
+    version_msg("\n");
     version_msg(_("  DEBUG BUILD"));
 #endif
 }
@@ -952,9 +1641,11 @@ version_msg(s)
 {
     int		len = (int)STRLEN(s);
 
-    if (len < (int)Columns && msg_col + len >= (int)Columns)
+    if (!got_int && len < (int)Columns && msg_col + len >= (int)Columns
+								&& *s != '\n')
 	msg_putchar('\n');
-    MSG_PUTS(s);
+    if (!got_int)
+	MSG_PUTS(s);
 }
 
 static void do_intro_line __ARGS((int row, char_u *mesg, int add_version, int attr));
@@ -971,6 +1662,7 @@ intro_message(colon)
     int		i;
     int		row;
     int		blanklines;
+    int		sponsor;
     char	*p;
     static char	*(lines[]) =
     {
@@ -988,7 +1680,7 @@ intro_message(colon)
 	"",
 	N_("type  :q<Enter>               to exit         "),
 	N_("type  :help<Enter>  or  <F1>  for on-line help"),
-	N_("type  :help version6<Enter>   for version info"),
+	N_("type  :help version7<Enter>   for version info"),
 	NULL,
 	"",
 	N_("Running in Vi compatible mode"),
@@ -1038,6 +1730,11 @@ intro_message(colon)
     if (blanklines < 0)
 	blanklines = 0;
 
+    /* Show the sponsor and register message one out of four times, the Uganda
+     * message two out of four times. */
+    sponsor = (int)time(NULL);
+    sponsor = ((sponsor & 2) == 0) - ((sponsor & 4) == 0);
+
     /* start displaying the message lines after half of the blank lines */
     row = blanklines / 2;
     if ((row >= 2 && Columns >= 50) || colon)
@@ -1055,6 +1752,19 @@ intro_message(colon)
 		    break;
 		continue;
 	    }
+	    if (sponsor != 0)
+	    {
+		if (strstr(p, "children") != NULL)
+		    p = sponsor < 0
+			? N_("Sponsor Vim development!")
+			: N_("Become a registered Vim user!");
+		else if (strstr(p, "iccf") != NULL)
+		    p = sponsor < 0
+			? N_("type  :help sponsor<Enter>    for information ")
+			: N_("type  :help register<Enter>   for information ");
+		else if (strstr(p, "Orphans") != NULL)
+		    p = N_("menu  Help->Sponsor/Register  for information    ");
+	    }
 	    if (*p != NUL)
 		do_intro_line(row, (char_u *)_(p), i == 2, 0);
 	    ++row;
@@ -1088,18 +1798,16 @@ do_intro_line(row, mesg, add_version, attr)
     int		col;
     char_u	*p;
     int		l;
-# define MODBY_LEN 150
-    char_u	buf[MODBY_LEN];
+    int		clen;
 #ifdef MODIFIED_BY
+# define MODBY_LEN 150
     char_u	modby[MODBY_LEN];
 
     if (*mesg == ' ')
     {
-	STRNCPY(modby, _("Modified by "), MODBY_LEN);
-	modby[MODBY_LEN - 1] = NUL;
+	vim_strncpy(modby, (char_u *)_("Modified by "), MODBY_LEN - 1);
 	l = STRLEN(modby);
-	STRNCPY(modby + l, MODIFIED_BY, MODBY_LEN - l);
-	modby[MODBY_LEN - 1] = NUL;
+	vim_strncpy(modby + l, (char_u *)MODIFIED_BY, MODBY_LEN - l - 1);
 	mesg = modby;
     }
 #endif
@@ -1133,12 +1841,22 @@ do_intro_line(row, mesg, add_version, attr)
     /* Split up in parts to highlight <> items differently. */
     for (p = mesg; *p != NUL; p += l)
     {
-	buf[0] = *p;
-	for (l = 1; p[l] != NUL && p[l] != '<' && p[l - 1] != '>'; ++l)
-	    buf[l] = p[l];
-	buf[l] = NUL;
-	screen_puts(buf, row, col, *buf == '<' ? hl_attr(HLF_8) : attr);
-	col += vim_strsize(buf);
+	clen = 0;
+	for (l = 0; p[l] != NUL
+			 && (l == 0 || (p[l] != '<' && p[l - 1] != '>')); ++l)
+	{
+#ifdef FEAT_MBYTE
+	    if (has_mbyte)
+	    {
+		clen += ptr2cells(p + l);
+		l += (*mb_ptr2len)(p + l) - 1;
+	    }
+	    else
+#endif
+		clen += byte2cells(p[l]);
+	}
+	screen_puts_len(p, l, row, col, *p == '<' ? hl_attr(HLF_8) : attr);
+	col += clen;
     }
 
     /* Add the version number to the version line. */
